from dataclasses import dataclass, field
from typing import ClassVar, Literal

from ..idl import IdlStruct, types
from ..std_msgs.msg import Header


@dataclass
class KeyValue(IdlStruct, typename="diagnostic_msgs/msg/KeyValue"):
    key: str = ""
    value: str = ""


@dataclass
class DiagnosticStatus(IdlStruct, typename="diagnostic_msgs/msg/DiagnosticStatus"):
    level: types.byte = 0
    name: str = ""
    message: str = ""
    hardware_id: str = ""
    values: types.sequence[KeyValue] = field(default_factory=list)
    OK: ClassVar[Literal[0]] = 0
    WARN: ClassVar[Literal[1]] = 1
    ERROR: ClassVar[Literal[2]] = 2
    STALE: ClassVar[Literal[3]] = 3

@dataclass
class DiagnosticArray(IdlStruct, typename="diagnostic_msgs/msg/DiagnosticArray"):
    header: Header = field(default_factory=Header)
    status: types.sequence[DiagnosticStatus] = field(default_factory=list)



from dataclasses import dataclass, field

from ..idl import IdlStruct, types


@dataclass
class UUID(IdlStruct, typename="unique_identifier_msgs/msg/UUID"):
    uuid: types.array[types.uint8, 16] = field(default_factory=lambda: [0] * 16)



from dataclasses import dataclass, field

from ..builtin_interfaces.msg import Time
from ..idl import IdlStruct, types


@dataclass
class Bool(IdlStruct, typename="std_msgs/msg/Bool"):
    data: bool = False


@dataclass
class Byte(IdlStruct, typename="std_msgs/msg/Byte"):
    data: types.byte = 0


@dataclass
class Char(IdlStruct, typename="std_msgs/msg/Char"):
    # type_description_interfaces.srv.GetTypeDescription_Response(
    #     successful=True,
    #     failure_reason="",
    #     type_description=type_description_interfaces.msg.TypeDescription(
    #         type_description=type_description_interfaces.msg.IndividualTypeDescription(
    #             type_name="std_msgs/msg/Char",
    #             fields=[
    #                 type_description_interfaces.msg.Field(
    #                     name="data",
    #                     type=type_description_interfaces.msg.FieldType(
    #                         type_id=3,
    #                         capacity=0,
    #                         string_capacity=0,
    #                         nested_type_name="",
    #                     ),
    #                     default_value="",
    #                 )
    #             ],
    #         ),
    #         referenced_type_descriptions=[],
    #     ),
    #     type_sources=[
    #         type_description_interfaces.msg.TypeSource(
    #             type_name="std_msgs/msg/Char",
    #             encoding="msg",
    #             raw_file_contents="# This was originally provided as an example message.\n# It is deprecated as of Foxy\n# It is recommended to create your own semantically meaningful message.\n# However if you would like to continue using this please use the equivalent in example_msgs.\n\nchar data",
    #         )
    #     ],
    #     extra_information=[],
    # )
    data: types.uint8 = 0


@dataclass
class ColorRGBA(IdlStruct, typename="std_msgs/msg/ColorRGBA"):
    r: types.float32 = 0.0
    g: types.float32 = 0.0
    b: types.float32 = 0.0
    a: types.float32 = 0.0


@dataclass
class Empty(IdlStruct, typename="std_msgs/msg/Empty"):
    # type_description_interfaces.srv.GetTypeDescription_Response(
    #     successful=True,
    #     failure_reason="",
    #     type_description=type_description_interfaces.msg.TypeDescription(
    #         type_description=type_description_interfaces.msg.IndividualTypeDescription(
    #             type_name="std_msgs/msg/Empty",
    #             fields=[
    #                 type_description_interfaces.msg.Field(
    #                     name="structure_needs_at_least_one_member",
    #                     type=type_description_interfaces.msg.FieldType(
    #                         type_id=3,
    #                         capacity=0,
    #                         string_capacity=0,
    #                         nested_type_name="",
    #                     ),
    #                     default_value="",
    #                 )
    #             ],
    #         ),
    #         referenced_type_descriptions=[],
    #     ),
    #     type_sources=[
    #         type_description_interfaces.msg.TypeSource(
    #             type_name="std_msgs/msg/Empty", encoding="msg", raw_file_contents=""
    #         )
    #     ],
    #     extra_information=[],
    # )
    # structure_needs_at_least_one_member: types.uint8 = 0
    pass

@dataclass
class Float32(IdlStruct, typename="std_msgs/msg/Float32"):
    data: types.float32 = 0.0


@dataclass
class Float64(IdlStruct, typename="std_msgs/msg/Float64"):
    data: types.float64 = 0.0


@dataclass
class Header(IdlStruct, typename="std_msgs/msg/Header"):
    stamp: Time = field(default_factory=Time)
    frame_id: str = ""


@dataclass
class Int16(IdlStruct, typename="std_msgs/msg/Int16"):
    data: types.int16 = 0


@dataclass
class Int32(IdlStruct, typename="std_msgs/msg/Int32"):
    data: types.int32 = 0


@dataclass
class Int64(IdlStruct, typename="std_msgs/msg/Int64"):
    data: types.int64 = 0


@dataclass
class Int8(IdlStruct, typename="std_msgs/msg/Int8"):
    data: types.int8 = 0


@dataclass
class MultiArrayDimension(IdlStruct, typename="std_msgs/msg/MultiArrayDimension"):
    label: str = ""
    size: types.uint32 = 0
    stride: types.uint32 = 0


@dataclass
class String(IdlStruct, typename="std_msgs/msg/String"):
    data: str = ""


@dataclass
class UInt16(IdlStruct, typename="std_msgs/msg/UInt16"):
    data: types.uint16 = 0


@dataclass
class UInt32(IdlStruct, typename="std_msgs/msg/UInt32"):
    data: types.uint32 = 0


@dataclass
class UInt64(IdlStruct, typename="std_msgs/msg/UInt64"):
    data: types.uint64 = 0


@dataclass
class UInt8(IdlStruct, typename="std_msgs/msg/UInt8"):
    data: types.uint8 = 0


@dataclass
class MultiArrayLayout(IdlStruct, typename="std_msgs/msg/MultiArrayLayout"):
    dim: types.sequence[MultiArrayDimension] = field(default_factory=list)
    data_offset: types.uint32 = 0


@dataclass
class ByteMultiArray(IdlStruct, typename="std_msgs/msg/ByteMultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.byte] = field(default_factory=list)


@dataclass
class Float32MultiArray(IdlStruct, typename="std_msgs/msg/Float32MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.float32] = field(default_factory=list)


@dataclass
class Float64MultiArray(IdlStruct, typename="std_msgs/msg/Float64MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.float64] = field(default_factory=list)


@dataclass
class Int16MultiArray(IdlStruct, typename="std_msgs/msg/Int16MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.int16] = field(default_factory=list)


@dataclass
class Int32MultiArray(IdlStruct, typename="std_msgs/msg/Int32MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.int32] = field(default_factory=list)


@dataclass
class Int64MultiArray(IdlStruct, typename="std_msgs/msg/Int64MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.int64] = field(default_factory=list)


@dataclass
class Int8MultiArray(IdlStruct, typename="std_msgs/msg/Int8MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.int8] = field(default_factory=list)


@dataclass
class UInt16MultiArray(IdlStruct, typename="std_msgs/msg/UInt16MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.uint16] = field(default_factory=list)


@dataclass
class UInt32MultiArray(IdlStruct, typename="std_msgs/msg/UInt32MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.uint32] = field(default_factory=list)


@dataclass
class UInt64MultiArray(IdlStruct, typename="std_msgs/msg/UInt64MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.uint64] = field(default_factory=list)


@dataclass
class UInt8MultiArray(IdlStruct, typename="std_msgs/msg/UInt8MultiArray"):
    layout: MultiArrayLayout = field(default_factory=MultiArrayLayout)
    data: types.sequence[types.uint8] = field(default_factory=list)



import inspect
import json
from typing import List, Type, get_type_hints

import pytest
from rclpy.serialization import deserialize_message, serialize_message
from utils import assert_msg_equal_as_lists, assert_strictly_eq

from ros2_pyterfaces import all_srvs, idl
from ros2_pyterfaces.all_srvs import (
    AddDiagnostics,
    AddDiagnostics_Request,
    AddDiagnostics_Response,
    GetTypeDescription,
    GetTypeDescription_Event,
    GetTypeDescription_Request,
    GetTypeDescription_Response,
    GetInteractiveMarkers_Response,
    GetPlan_Request,
    LoadMap,
    LoadMap_Response,
    SelfTest_Response,
    SetBool,
    SetBool_Event,
    SetBool_Request,
    SetBool_Response,
    SetCameraInfo_Response,
    Trigger,
)
from ros2_pyterfaces.diagnostic_msgs.msg import DiagnosticStatus
from ros2_pyterfaces.geometry_msgs.msg import Point, Pose, PoseStamped, Quaternion
from ros2_pyterfaces.nav_msgs.msg import MapMetaData, OccupancyGrid
from ros2_pyterfaces.sensor_msgs.msg import CameraInfo
from ros2_pyterfaces.std_msgs.msg import Header
from ros2_pyterfaces.type_description_interfaces.msg import (
    Field,
    FieldType,
    IndividualTypeDescription,
    KeyValue,
    TypeDescription,
    TypeSource,
)
from ros2_pyterfaces.visualization_msgs.msg import InteractiveMarker

SERVICE_TYPES = [
    obj
    for obj in vars(all_srvs).values()
    if inspect.isclass(obj) and idl.is_service_type(obj)
]
REQUEST_RESPONSE_TYPES: List[Type[idl.IdlStruct]] = [
    obj
    for obj in vars(all_srvs).values()
    if inspect.isclass(obj)
    and issubclass(obj, idl.IdlStruct)
    and obj is not idl.IdlStruct
    and not idl.is_service_type(obj)
]

VALUES = [
    SetBool_Request(data=True),
    SetBool_Response(success=True, message="done"),
    AddDiagnostics_Request(load_namespace="/robot/diagnostics"),
    AddDiagnostics_Response(success=True, message="loaded"),
    SelfTest_Response(
        id="motor-check",
        passed=1,
        status=[
            DiagnosticStatus(
                level=DiagnosticStatus.OK,
                name="motor",
                message="ok",
            )
        ],
    ),
    GetPlan_Request(
        start=PoseStamped(
            header=Header(frame_id="map"),
            pose=Pose(position=Point(1, 2, 0), orientation=Quaternion(1, 0, 0, 0)),
        ),
        goal=PoseStamped(
            header=Header(frame_id="map"),
            pose=Pose(position=Point(3, 4, 0), orientation=Quaternion(1, 0, 0, 0)),
        ),
        tolerance=1,
    ),
    LoadMap_Response(
        map=OccupancyGrid(info=MapMetaData(width=2, height=2), data=[0, 100, -1, 0]),
        result=LoadMap_Response.RESULT_SUCCESS,
    ),
    SetCameraInfo_Response(success=True, status_message="stored"),
    GetTypeDescription_Request(
        type_name="std_msgs/msg/String",
        type_hash="RIHS01_123",
        include_type_sources=False,
    ),
    GetTypeDescription_Response(
        successful=True,
        type_description=TypeDescription(
            type_description=IndividualTypeDescription(
                type_name="std_msgs/msg/String",
                fields=[
                    Field(
                        name="data",
                        type=FieldType(type_id=FieldType.FIELD_TYPE_STRING),
                        default_value="",
                    )
                ],
            )
        ),
        type_sources=[
            TypeSource(
                type_name="std_msgs/msg/String",
                encoding="msg",
                raw_file_contents="string data\n",
            )
        ],
        extra_information=[KeyValue(key="schema", value="test")],
    ),
    GetInteractiveMarkers_Response(
        sequence_number=3,
        markers=[InteractiveMarker(name="control")],
    ),
]


@pytest.mark.parametrize("srv_type", SERVICE_TYPES)
def test_service_wrapper_matches_ros(srv_type: Type[idl.IdlStruct]):
    ros_srv_type = srv_type.get_ros_type()
    type_hints = get_type_hints(srv_type)

    assert srv_type.to_ros_type() is ros_srv_type
    assert type_hints["request_message"].get_ros_type() is ros_srv_type.Request
    assert type_hints["response_message"].get_ros_type() is ros_srv_type.Response


@pytest.mark.parametrize("msg_type", REQUEST_RESPONSE_TYPES)
def test_service_messages_to_ros_defaults(msg_type: Type[idl.IdlStruct]):
    ros_msg_type = msg_type.get_ros_type()
    ros_msg = msg_type().to_ros()

    assert isinstance(ros_msg, ros_msg_type)
    assert_msg_equal_as_lists(ros_msg, ros_msg_type())


@pytest.mark.parametrize("msg_type", REQUEST_RESPONSE_TYPES)
def test_service_messages_from_ros_defaults(msg_type: Type[idl.IdlStruct]):
    ros_msg_type = msg_type.get_ros_type()
    idl_msg = msg_type.from_ros(ros_msg_type())

    assert_strictly_eq(idl_msg, msg_type())


@pytest.mark.parametrize("msg_type", REQUEST_RESPONSE_TYPES)
def test_service_deserialize(msg_type: Type[idl.IdlStruct]):
    ros_msg_type = msg_type.get_ros_type()
    idl_from_ros = msg_type.deserialize(serialize_message(ros_msg_type()))

    assert_strictly_eq(idl_from_ros, msg_type())


@pytest.mark.parametrize("msg_type", REQUEST_RESPONSE_TYPES)
def test_service_serialize(msg_type: Type[idl.IdlStruct]):
    ros_msg_type = msg_type.get_ros_type()
    ros_from_idl = deserialize_message(msg_type().serialize(), ros_msg_type)

    assert_msg_equal_as_lists(ros_from_idl, ros_msg_type())


@pytest.mark.parametrize("msg", VALUES)
def test_service_messages_to_from_ros_values(msg: idl.IdlStruct):
    ros_msg = msg.to_ros()

    assert isinstance(ros_msg, type(msg).get_ros_type())
    assert_strictly_eq(type(msg).from_ros(ros_msg), msg)


@pytest.mark.parametrize("msg", VALUES)
def test_service_deserialize_values(msg: idl.IdlStruct):
    idl_from_ros = type(msg).deserialize(serialize_message(msg.to_ros()))

    assert_strictly_eq(idl_from_ros, msg)


@pytest.mark.parametrize("msg", VALUES)
def test_service_serialize_values(msg: idl.IdlStruct):
    ros_msg_type = msg.get_ros_type()
    ros_from_idl = deserialize_message(msg.serialize(), ros_msg_type)

    assert_strictly_eq(type(msg).from_ros(ros_from_idl), msg)
    assert_strictly_eq(type(msg).deserialize(serialize_message(ros_from_idl)), msg)


def test_service_wrapper_request_response_links():
    setbool_hints = get_type_hints(SetBool)
    load_map_hints = get_type_hints(LoadMap)
    get_type_description_hints = get_type_hints(GetTypeDescription)

    assert SetBool.Request is SetBool_Request
    assert SetBool.Response is SetBool_Response
    assert SetBool.Event is SetBool_Event
    assert LoadMap.Response is LoadMap_Response
    assert GetTypeDescription.Event is GetTypeDescription_Event
    assert setbool_hints["request_message"] is SetBool_Request
    assert setbool_hints["response_message"] is SetBool_Response
    assert setbool_hints["event_message"].get_type_name() == "std_srvs/srv/SetBool_Event"
    assert load_map_hints["response_message"] is LoadMap_Response
    assert get_type_description_hints["request_message"] is GetTypeDescription_Request
    assert (
        get_type_description_hints["response_message"] is GetTypeDescription_Response
    )
    assert GetTypeDescription.Request is GetTypeDescription_Request
    assert GetTypeDescription.Response is GetTypeDescription_Response


def test_setbool_readme_service_usage():
    some_request: SetBool_Request = SetBool.Request(data=True)
    some_response: SetBool_Response = SetBool.Response(
        success=True,
        message="yey",
    )

    ros_srv_type = SetBool.to_ros_type()

    ros_request = some_request.to_ros()
    request_again = SetBool.Request.from_ros(ros_request)

    ros_response = some_response.to_ros()
    response_again = SetBool.Response.from_ros(ros_response)

    assert ros_srv_type is SetBool.get_ros_type()
    assert isinstance(ros_request, ros_srv_type.Request)
    assert isinstance(ros_response, ros_srv_type.Response)
    assert_strictly_eq(request_again, some_request)
    assert_strictly_eq(response_again, some_response)


def test_all_srvs_exposes_event_types():
    assert issubclass(GetTypeDescription_Event, idl.IdlStruct)


def test_get_type_description_request_defaults_to_include_type_sources():
    request = GetTypeDescription_Request()

    assert request.include_type_sources is True
    assert request.to_ros().include_type_sources is True


def test_trigger_request_type_description_uses_placeholder_member():
    type_description = json.loads(Trigger.json_type_description())
    referenced = {
        desc["type_name"]: desc for desc in type_description["referenced_type_descriptions"]
    }
    trigger_request = referenced["std_srvs/srv/Trigger_Request"]

    assert trigger_request["fields"] == [
        {
            "name": "structure_needs_at_least_one_member",
            "type": {
                "type_id": 3,
                "capacity": 0,
                "string_capacity": 0,
                "nested_type_name": "",
            },
            "default_value": "",
        }
    ]



from __future__ import annotations

import copy
import hashlib
import json
import sys
import typing as t
from collections.abc import Mapping
from typing import Any, Dict

from cyclonedds_idl import IdlStruct, types  # I wanna delete this and depend on .idl


def ros2_type_hash_from_json(type_description_json: str) -> "hashlib._Hash":
    """
    Compute the raw RIHS01 hash from type description JSON.
    """

    raw = json.loads(type_description_json)

    if "successful" in raw:
        if not raw.get("successful", False):
            raise ValueError(
                f"GetTypeDescription response is not successful: {raw.get('failure_reason', '')}"
            )
        raw = raw["type_description"]

    if not isinstance(raw, dict):
        raise TypeError("Top-level JSON must decode to an object")

    if "type_description" not in raw or "referenced_type_descriptions" not in raw:
        raise ValueError(
            "Expected a TypeDescription JSON object with keys "
            "'type_description' and 'referenced_type_descriptions'"
        )

    def normalize_field_type(ft: dict[str, Any]) -> dict[str, Any]:
        return {
            "type_id": ft["type_id"],
            "capacity": ft["capacity"],
            "string_capacity": ft["string_capacity"],
            "nested_type_name": ft["nested_type_name"],
        }

    def normalize_field(field: dict[str, Any]) -> dict[str, Any]:
        return {
            "name": field["name"],
            "type": normalize_field_type(field["type"]),
        }

    def normalize_individual(td: dict[str, Any]) -> dict[str, Any]:
        return {
            "type_name": td["type_name"],
            "fields": [normalize_field(f) for f in td["fields"]],
        }

    hashable = {
        "type_description": normalize_individual(raw["type_description"]),
        "referenced_type_descriptions": [
            normalize_individual(td)
            for td in sorted(
                raw["referenced_type_descriptions"],
                key=lambda td: td["type_name"],
            )
        ],
    }

    hashable_repr = json.dumps(
        hashable,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=False,
        indent=None,
        separators=(", ", ": "),
        sort_keys=False,
    )

    return hashlib.sha256(hashable_repr.encode("utf-8"))


_ROS_SCALAR_TYPE_IDS = {
    "nested": 1,
    "int8": 2,
    "uint8": 3,
    "int16": 4,
    "uint16": 5,
    "int32": 6,
    "uint32": 7,
    "int64": 8,
    "uint64": 9,
    "float32": 10,
    "float64": 11,
    "char": 13,  # replaced by uint8?
    "wchar": 14,
    "bool": 15,
    "byte": 16,
    "string": 17,
    "wstring": 18,
    "fixed_string": 19,
    "fixed_wstring": 20,
    "bounded_string": 21,
    "bounded_wstring": 22,
}

_ARRAY_OFFSET = 48
_BOUNDED_SEQUENCE_OFFSET = 96
_UNBOUNDED_SEQUENCE_OFFSET = 144

_CYCLONE_PRIMITIVE_TAG_TO_ROS = {
    "int8": _ROS_SCALAR_TYPE_IDS["int8"],
    "uint8": _ROS_SCALAR_TYPE_IDS["uint8"],
    "int16": _ROS_SCALAR_TYPE_IDS["int16"],
    "uint16": _ROS_SCALAR_TYPE_IDS["uint16"],
    "int32": _ROS_SCALAR_TYPE_IDS["int32"],
    "uint32": _ROS_SCALAR_TYPE_IDS["uint32"],
    "int64": _ROS_SCALAR_TYPE_IDS["int64"],
    "uint64": _ROS_SCALAR_TYPE_IDS["uint64"],
    "float32": _ROS_SCALAR_TYPE_IDS["float32"],
    "float64": _ROS_SCALAR_TYPE_IDS["float64"],
    "char": _ROS_SCALAR_TYPE_IDS["char"],
    "wchar": _ROS_SCALAR_TYPE_IDS["wchar"],
    "byte": _ROS_SCALAR_TYPE_IDS["byte"],
}


def _placeholder_field() -> dict[str, Any]:
    """
    Build the placeholder field used for empty structures.
    """
    return {
        "name": "structure_needs_at_least_one_member",
        "type": {
            "type_id": _ROS_SCALAR_TYPE_IDS["uint8"],
            "capacity": 0,
            "string_capacity": 0,
            "nested_type_name": "",
        },
        "default_value": "",
    }


def cyclonedds_struct_to_ros_type_description_json(
    cls: type[IdlStruct],
    *,
    root_type_name: str | None = None,
    type_name_overrides: Mapping[type, str] | None = None,
    indent: int = 2,
) -> str:
    """
    Convert an IDL struct type to ROS 2 type description JSON.

    Args:
        cls: IDL struct type.
        root_type_name: Optional root ROS type name.
        type_name_overrides: Optional nested type name overrides.
        indent: JSON indentation level.

    Returns:
        Type description JSON.
    """
    if not isinstance(cls, type) or not issubclass(cls, IdlStruct):
        raise TypeError("cls must be a cyclonedds.idl.IdlStruct subclass")

    types_map: Mapping[type, str] = (
        dict(type_name_overrides) if type_name_overrides is not None else dict()
    )

    def resolve_ros_type_name(tp_cls: type, *, is_root: bool = False) -> str:
        nonlocal types_map, root_type_name
        if tp_cls in types_map:
            return types_map[tp_cls]

        if is_root and root_type_name:
            return root_type_name

        name = getattr(tp_cls, "__idl_typename__", None)
        if isinstance(name, str) and "/" in name:
            return name

        raise ValueError(
            f"No ROS type name available for {tp_cls.__module__}.{tp_cls.__qualname__}. "
            f"Pass root_type_name=... for the root class, and/or type_name_overrides={{ThatClass: 'pkg/msg/Name'}} "
            f"for nested classes."
        )

    def unwrap_annotated(tp: t.Any) -> tuple[t.Any, list[t.Any]]:
        metadata: list[t.Any] = []
        while t.get_origin(tp) is t.Annotated:
            args = t.get_args(tp)
            tp = args[0]
            metadata.extend(args[1:])
        return tp, metadata

    def metadata_by_class_name(metadata: list[t.Any], class_name: str) -> t.Any | None:
        for item in metadata:
            if item.__class__.__name__ == class_name:
                return item
        return None

    def primitive_type_id_from_python_type(
        base: t.Any, metadata: list[t.Any]
    ) -> int | None:
        for item in metadata:
            if isinstance(item, str) and item in _CYCLONE_PRIMITIVE_TAG_TO_ROS:
                return _CYCLONE_PRIMITIVE_TAG_TO_ROS[item]
        if base is str:
            return _ROS_SCALAR_TYPE_IDS["string"]
        if base is bool:
            return _ROS_SCALAR_TYPE_IDS["bool"]
        if base is int:
            return _ROS_SCALAR_TYPE_IDS["int64"]
        if base is float:
            return _ROS_SCALAR_TYPE_IDS["float64"]
        return None

    def field_type_from_annotation(tp: t.Any) -> tuple[dict[str, t.Any], set[type]]:
        base, metadata = unwrap_annotated(tp)

        td = metadata_by_class_name(metadata, "typedef")
        if td is not None:
            return field_type_from_annotation(td.subtype)

        arr = metadata_by_class_name(metadata, "array")
        seq = metadata_by_class_name(metadata, "sequence")
        bstr = metadata_by_class_name(metadata, "bounded_str")

        if arr is not None and seq is not None:
            raise TypeError(
                f"Unsupported type annotation with both array and sequence metadata: {tp!r}"
            )

        if bstr is not None:
            return (
                {
                    "type_id": _ROS_SCALAR_TYPE_IDS["bounded_string"],
                    "capacity": 0,
                    "string_capacity": int(bstr.max_length),
                    "nested_type_name": "",
                },
                set(),
            )

        if arr is not None:
            elem_ft, refs = field_type_from_annotation(arr.subtype)
            return (
                {
                    "type_id": int(elem_ft["type_id"]) + _ARRAY_OFFSET,
                    "capacity": int(arr.length),
                    "string_capacity": int(elem_ft["string_capacity"]),
                    "nested_type_name": elem_ft["nested_type_name"],
                },
                refs,
            )

        if seq is not None:
            elem_ft, refs = field_type_from_annotation(seq.subtype)
            if seq.max_length is None:
                type_id = int(elem_ft["type_id"]) + _UNBOUNDED_SEQUENCE_OFFSET
                capacity = 0
            else:
                type_id = int(elem_ft["type_id"]) + _BOUNDED_SEQUENCE_OFFSET
                capacity = int(seq.max_length)
            return (
                {
                    "type_id": type_id,
                    "capacity": capacity,
                    "string_capacity": int(elem_ft["string_capacity"]),
                    "nested_type_name": elem_ft["nested_type_name"],
                },
                refs,
            )

        prim = primitive_type_id_from_python_type(base, metadata)
        if prim is not None:
            return (
                {
                    "type_id": prim,
                    "capacity": 0,
                    "string_capacity": 0,
                    "nested_type_name": "",
                },
                set(),
            )

        if isinstance(base, type) and issubclass(base, IdlStruct):
            return (
                {
                    "type_id": _ROS_SCALAR_TYPE_IDS["nested"],
                    "capacity": 0,
                    "string_capacity": 0,
                    "nested_type_name": resolve_ros_type_name(base, is_root=False),
                },
                {base},
            )

        raise TypeError(f"Unsupported field type annotation: {tp!r}")

    seen: dict[type, dict[str, t.Any]] = {}

    def build_individual_type(
        tp_cls: type[IdlStruct], *, is_root: bool = False
    ) -> None:
        if tp_cls in seen:
            return

        module_globals = vars(sys.modules[tp_cls.__module__])

        hints = t.get_type_hints(
            tp_cls,
            globalns=module_globals,
            localns=module_globals,
            include_extras=True,
        )
        raw_annotations = getattr(tp_cls, "__annotations__", {})
        field_annotations = getattr(tp_cls, "__idl_field_annotations__", {})

        referenced: set[type] = set()
        fields: list[dict[str, t.Any]] = []

        for py_field_name in raw_annotations:
            field_tp = hints.get(py_field_name, raw_annotations[py_field_name])
            field_type, refs = field_type_from_annotation(field_tp)
            referenced |= refs

            ros_field_name = field_annotations.get(py_field_name, {}).get(
                "name", py_field_name
            )

            fields.append(
                {
                    "name": ros_field_name,
                    "type": field_type,
                    "default_value": "",
                }
            )

        if len(fields) == 0:
            fields = [_placeholder_field()]

        seen[tp_cls] = {
            "type_name": resolve_ros_type_name(tp_cls, is_root=is_root),
            "fields": fields,
        }

        for ref_cls in referenced:
            build_individual_type(ref_cls, is_root=False)

    build_individual_type(cls, is_root=True)

    root = seen[cls]
    referenced = [desc for c, desc in seen.items() if c is not cls]
    referenced.sort(key=lambda d: d["type_name"])

    out = {
        "type_description": root,
        "referenced_type_descriptions": referenced,
    }
    return json.dumps(out, indent=indent)



import time
from dataclasses import dataclass, field
from typing import Tuple, Type, get_type_hints

import asyncio_for_robotics.ros2 as afor
import pytest
import rclpy

from ros2_pyterfaces import all_msgs, all_srvs, idl
from ros2_pyterfaces.all_srvs import (
    SetBool,
    SetBool_Event,
    SetBool_Request,
    SetBool_Response,
    Trigger,
)
from ros2_pyterfaces.sensor_msgs.srv import SetCameraInfo

HelperSetBool = idl.make_idl_service(
    SetBool_Request,
    SetBool_Response,
    event_type=SetBool_Event,
)

TYPES_TO_HASH: dict[type[idl.IdlStruct], str] = {
    SetCameraInfo: "RIHS01_a10cca5d33dc637c8d49db50ab288701a3592bb9cd854f2f16a0659613b68984",
    Trigger: "RIHS01_eeff2cd6fa5ad9d27cdf4dec64818317839b62f212a91e6b5304b634b2062c5f",
    SetBool: "RIHS01_abe9e4bb6b41b40e6789712c00ec8871923e089af3f667a79992a428cff2da0a",
    HelperSetBool: "RIHS01_abe9e4bb6b41b40e6789712c00ec8871923e089af3f667a79992a428cff2da0a",
}


def debug(my_type: Type[idl.IdlStruct]) -> Tuple[str, str]:
    print(Trigger.json_type_description())
    print(Trigger.hash_rihs01())
    print(Trigger().request_message.hash_rihs01())
    print(Trigger().response_message.hash_rihs01())
    print(Trigger().event_message.hash_rihs01())
    time.sleep(2)
    cli = afor.Client(
        my_type.get_ros_type(), f"/get_client_hash/{my_type.get_type_name()}"
    )
    time.sleep(2)
    srv = afor.Server(
        my_type.get_ros_type(), f"/get_client_hash/{my_type.get_type_name()}"
    )
    time.sleep(100000)
    try:
        pytest.skip()
        return info[0]["topic_type"], info[0]["topic_type_hash"]["value"]
    finally:
        srv.close()
        cli.close()


@pytest.fixture(scope="module")
def init():
    rclpy.init()
    ses = afor.auto_session()
    yield
    ses.close()
    try:
        rclpy.shutdown()
    except:
        pass


@pytest.mark.parametrize("my_type", TYPES_TO_HASH.keys())
async def test_hashes(my_type: type[idl.IdlStruct], init):
    # debug(my_type)
    assert my_type.hash_rihs01() == TYPES_TO_HASH[my_type]



from dataclasses import dataclass, field

from .. import idl
from ..service_msgs.msg import ServiceEventInfo
from ..sensor_msgs.msg import CameraInfo

@dataclass
class SetCameraInfo_Request(
    idl.IdlStruct, typename="sensor_msgs/srv/SetCameraInfo_Request"
):
    camera_info: CameraInfo = field(default_factory=CameraInfo)


@dataclass
class SetCameraInfo_Response(
    idl.IdlStruct, typename="sensor_msgs/srv/SetCameraInfo_Response"
):
    success: bool = False
    status_message: str = ""


@dataclass
class SetCameraInfo_Event(
    idl.IdlStruct,
    typename="sensor_msgs/srv/SetCameraInfo_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[SetCameraInfo_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[SetCameraInfo_Response, 1] = field(default_factory=list)


SetCameraInfo: idl.IdlServiceType[
    SetCameraInfo_Request,
    SetCameraInfo_Response,
    SetCameraInfo_Event,
] = idl.make_idl_service(
    SetCameraInfo_Request,
    SetCameraInfo_Response,
    event_type=SetCameraInfo_Event,
)



from dataclasses import dataclass, field

from .. import idl
from .msg import State, Transition, TransitionDescription


@dataclass
class ChangeState_Request(idl.IdlStruct, typename="lifecycle_msgs/srv/ChangeState_Request"):
    transition: Transition = field(default_factory=Transition)


@dataclass
class ChangeState_Response(idl.IdlStruct, typename="lifecycle_msgs/srv/ChangeState_Response"):
    success: bool = False


ChangeState = idl.make_idl_service(ChangeState_Request, ChangeState_Response)
ChangeState_Event = ChangeState.Event


@dataclass
class GetAvailableStates_Request(
    idl.IdlStruct, typename="lifecycle_msgs/srv/GetAvailableStates_Request"
):
    pass


@dataclass
class GetAvailableStates_Response(
    idl.IdlStruct, typename="lifecycle_msgs/srv/GetAvailableStates_Response"
):
    available_states: idl.types.sequence[State] = field(default_factory=list)


GetAvailableStates = idl.make_idl_service(
    GetAvailableStates_Request, GetAvailableStates_Response
)
GetAvailableStates_Event = GetAvailableStates.Event


@dataclass
class GetAvailableTransitions_Request(
    idl.IdlStruct, typename="lifecycle_msgs/srv/GetAvailableTransitions_Request"
):
    pass


@dataclass
class GetAvailableTransitions_Response(
    idl.IdlStruct, typename="lifecycle_msgs/srv/GetAvailableTransitions_Response"
):
    available_transitions: idl.types.sequence[TransitionDescription] = field(
        default_factory=list
    )


GetAvailableTransitions = idl.make_idl_service(
    GetAvailableTransitions_Request, GetAvailableTransitions_Response
)
GetAvailableTransitions_Event = GetAvailableTransitions.Event


@dataclass
class GetState_Request(idl.IdlStruct, typename="lifecycle_msgs/srv/GetState_Request"):
    pass


@dataclass
class GetState_Response(idl.IdlStruct, typename="lifecycle_msgs/srv/GetState_Response"):
    current_state: State = field(default_factory=State)


GetState = idl.make_idl_service(GetState_Request, GetState_Response)
GetState_Event = GetState.Event



from dataclasses import dataclass, field
from typing import ClassVar, Literal

import numpy as np

from ..geometry_msgs.msg import Point, Polygon
from ..idl import IdlStruct, types


@dataclass
class MeshTriangle(IdlStruct, typename="shape_msgs/msg/MeshTriangle"):
    vertex_indices: types.array[types.uint32, 3] = field(
        default_factory=lambda: np.zeros(3, dtype=np.uint32)
    )


@dataclass
class Plane(IdlStruct, typename="shape_msgs/msg/Plane"):
    coef: types.array[types.float64, 4] = field(default_factory=lambda: np.zeros(4))


@dataclass
class SolidPrimitive(IdlStruct, typename="shape_msgs/msg/SolidPrimitive"):
    BOX: ClassVar[Literal[1]] = 1
    SPHERE: ClassVar[Literal[2]] = 2
    CYLINDER: ClassVar[Literal[3]] = 3
    CONE: ClassVar[Literal[4]] = 4
    PRISM: ClassVar[Literal[5]] = 5
    BOX_X: ClassVar[Literal[0]] = 0
    BOX_Y: ClassVar[Literal[1]] = 1
    BOX_Z: ClassVar[Literal[2]] = 2
    SPHERE_RADIUS: ClassVar[Literal[0]] = 0
    CYLINDER_HEIGHT: ClassVar[Literal[0]] = 0
    CYLINDER_RADIUS: ClassVar[Literal[1]] = 1
    CONE_HEIGHT: ClassVar[Literal[0]] = 0
    CONE_RADIUS: ClassVar[Literal[1]] = 1
    PRISM_HEIGHT: ClassVar[Literal[0]] = 0
    type: types.uint8 = 0
    dimensions: types.sequence[types.float64, 3] = field(default_factory=list)
    polygon: Polygon = field(default_factory=Polygon)


@dataclass
class Mesh(IdlStruct, typename="shape_msgs/msg/Mesh"):
    triangles: types.sequence[MeshTriangle] = field(default_factory=list)
    vertices: types.sequence[Point] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from ..idl import IdlStruct, types


@dataclass
class State(IdlStruct, typename="lifecycle_msgs/msg/State"):
    PRIMARY_STATE_UNKNOWN: ClassVar[Literal[0]] = 0
    PRIMARY_STATE_UNCONFIGURED: ClassVar[Literal[1]] = 1
    PRIMARY_STATE_INACTIVE: ClassVar[Literal[2]] = 2
    PRIMARY_STATE_ACTIVE: ClassVar[Literal[3]] = 3
    PRIMARY_STATE_FINALIZED: ClassVar[Literal[4]] = 4
    TRANSITION_STATE_CONFIGURING: ClassVar[Literal[10]] = 10
    TRANSITION_STATE_CLEANINGUP: ClassVar[Literal[11]] = 11
    TRANSITION_STATE_SHUTTINGDOWN: ClassVar[Literal[12]] = 12
    TRANSITION_STATE_ACTIVATING: ClassVar[Literal[13]] = 13
    TRANSITION_STATE_DEACTIVATING: ClassVar[Literal[14]] = 14
    TRANSITION_STATE_ERRORPROCESSING: ClassVar[Literal[15]] = 15

    id: types.uint8 = 0
    label: str = ""


@dataclass
class Transition(IdlStruct, typename="lifecycle_msgs/msg/Transition"):
    TRANSITION_CREATE: ClassVar[Literal[0]] = 0
    TRANSITION_CONFIGURE: ClassVar[Literal[1]] = 1
    TRANSITION_CLEANUP: ClassVar[Literal[2]] = 2
    TRANSITION_ACTIVATE: ClassVar[Literal[3]] = 3
    TRANSITION_DEACTIVATE: ClassVar[Literal[4]] = 4
    TRANSITION_UNCONFIGURED_SHUTDOWN: ClassVar[Literal[5]] = 5
    TRANSITION_INACTIVE_SHUTDOWN: ClassVar[Literal[6]] = 6
    TRANSITION_ACTIVE_SHUTDOWN: ClassVar[Literal[7]] = 7
    TRANSITION_DESTROY: ClassVar[Literal[8]] = 8
    TRANSITION_ON_CONFIGURE_SUCCESS: ClassVar[Literal[10]] = 10
    TRANSITION_ON_CONFIGURE_FAILURE: ClassVar[Literal[11]] = 11
    TRANSITION_ON_CONFIGURE_ERROR: ClassVar[Literal[12]] = 12
    TRANSITION_ON_CLEANUP_SUCCESS: ClassVar[Literal[20]] = 20
    TRANSITION_ON_CLEANUP_FAILURE: ClassVar[Literal[21]] = 21
    TRANSITION_ON_CLEANUP_ERROR: ClassVar[Literal[22]] = 22
    TRANSITION_ON_ACTIVATE_SUCCESS: ClassVar[Literal[30]] = 30
    TRANSITION_ON_ACTIVATE_FAILURE: ClassVar[Literal[31]] = 31
    TRANSITION_ON_ACTIVATE_ERROR: ClassVar[Literal[32]] = 32
    TRANSITION_ON_DEACTIVATE_SUCCESS: ClassVar[Literal[40]] = 40
    TRANSITION_ON_DEACTIVATE_FAILURE: ClassVar[Literal[41]] = 41
    TRANSITION_ON_DEACTIVATE_ERROR: ClassVar[Literal[42]] = 42
    TRANSITION_ON_SHUTDOWN_SUCCESS: ClassVar[Literal[50]] = 50
    TRANSITION_ON_SHUTDOWN_FAILURE: ClassVar[Literal[51]] = 51
    TRANSITION_ON_SHUTDOWN_ERROR: ClassVar[Literal[52]] = 52
    TRANSITION_ON_ERROR_SUCCESS: ClassVar[Literal[60]] = 60
    TRANSITION_ON_ERROR_FAILURE: ClassVar[Literal[61]] = 61
    TRANSITION_ON_ERROR_ERROR: ClassVar[Literal[62]] = 62
    TRANSITION_CALLBACK_SUCCESS: ClassVar[Literal[97]] = 97
    TRANSITION_CALLBACK_FAILURE: ClassVar[Literal[98]] = 98
    TRANSITION_CALLBACK_ERROR: ClassVar[Literal[99]] = 99

    id: types.uint8 = 0
    label: str = ""


@dataclass
class TransitionDescription(
    IdlStruct, typename="lifecycle_msgs/msg/TransitionDescription"
):
    transition: Transition = field(default_factory=Transition)
    start_state: State = field(default_factory=State)
    goal_state: State = field(default_factory=State)


@dataclass
class TransitionEvent(IdlStruct, typename="lifecycle_msgs/msg/TransitionEvent"):
    timestamp: types.uint64 = 0
    transition: Transition = field(default_factory=Transition)
    start_state: State = field(default_factory=State)
    goal_state: State = field(default_factory=State)



from dataclasses import dataclass, field

import numpy as np

from ..idl import IdlStruct, types
from ..std_msgs.msg import Header


@dataclass
class Point(IdlStruct, typename="geometry_msgs/msg/Point"):
    x: types.float64 = 0.0
    y: types.float64 = 0.0
    z: types.float64 = 0.0


@dataclass
class Point32(IdlStruct, typename="geometry_msgs/msg/Point32"):
    x: types.float32 = 0.0
    y: types.float32 = 0.0
    z: types.float32 = 0.0


@dataclass
class Quaternion(IdlStruct, typename="geometry_msgs/msg/Quaternion"):
    x: types.float64 = 0.0
    y: types.float64 = 0.0
    z: types.float64 = 0.0
    w: types.float64 = 1.0


@dataclass
class Vector3(IdlStruct, typename="geometry_msgs/msg/Vector3"):
    x: types.float64 = 0.0
    y: types.float64 = 0.0
    z: types.float64 = 0.0


@dataclass
class Accel(IdlStruct, typename="geometry_msgs/msg/Accel"):
    linear: Vector3 = field(default_factory=Vector3)
    angular: Vector3 = field(default_factory=Vector3)


@dataclass
class Inertia(IdlStruct, typename="geometry_msgs/msg/Inertia"):
    m: types.float64 = 0.0
    com: Vector3 = field(default_factory=Vector3)
    ixx: types.float64 = 0.0
    ixy: types.float64 = 0.0
    ixz: types.float64 = 0.0
    iyy: types.float64 = 0.0
    iyz: types.float64 = 0.0
    izz: types.float64 = 0.0


@dataclass
class PointStamped(IdlStruct, typename="geometry_msgs/msg/PointStamped"):
    header: Header = field(default_factory=Header)
    point: Point = field(default_factory=Point)


@dataclass
class Polygon(IdlStruct, typename="geometry_msgs/msg/Polygon"):
    points: types.sequence[Point32] = field(default_factory=list)


@dataclass
class Pose(IdlStruct, typename="geometry_msgs/msg/Pose"):
    position: Point = field(default_factory=Point)
    orientation: Quaternion = field(default_factory=Quaternion)


@dataclass
class QuaternionStamped(IdlStruct, typename="geometry_msgs/msg/QuaternionStamped"):
    header: Header = field(default_factory=Header)
    quaternion: Quaternion = field(default_factory=Quaternion)


@dataclass
class Transform(IdlStruct, typename="geometry_msgs/msg/Transform"):
    translation: Vector3 = field(default_factory=Vector3)
    rotation: Quaternion = field(default_factory=Quaternion)


@dataclass
class Twist(IdlStruct, typename="geometry_msgs/msg/Twist"):
    linear: Vector3 = field(default_factory=Vector3)
    angular: Vector3 = field(default_factory=Vector3)


@dataclass
class Vector3Stamped(IdlStruct, typename="geometry_msgs/msg/Vector3Stamped"):
    header: Header = field(default_factory=Header)
    vector: Vector3 = field(default_factory=Vector3)


@dataclass
class Wrench(IdlStruct, typename="geometry_msgs/msg/Wrench"):
    force: Vector3 = field(default_factory=Vector3)
    torque: Vector3 = field(default_factory=Vector3)


@dataclass
class AccelStamped(IdlStruct, typename="geometry_msgs/msg/AccelStamped"):
    header: Header = field(default_factory=Header)
    accel: Accel = field(default_factory=Accel)


@dataclass
class AccelWithCovariance(IdlStruct, typename="geometry_msgs/msg/AccelWithCovariance"):
    accel: Accel = field(default_factory=Accel)
    covariance: types.array[types.float64, 36] = field(
        default_factory=lambda: np.zeros(36)
    )


@dataclass
class InertiaStamped(IdlStruct, typename="geometry_msgs/msg/InertiaStamped"):
    header: Header = field(default_factory=Header)
    inertia: Inertia = field(default_factory=Inertia)


@dataclass
class PolygonInstance(IdlStruct, typename="geometry_msgs/msg/PolygonInstance"):
    polygon: Polygon = field(default_factory=Polygon)
    id: types.int64 = 0


@dataclass
class PolygonStamped(IdlStruct, typename="geometry_msgs/msg/PolygonStamped"):
    header: Header = field(default_factory=Header)
    polygon: Polygon = field(default_factory=Polygon)


@dataclass
class PoseArray(IdlStruct, typename="geometry_msgs/msg/PoseArray"):
    header: Header = field(default_factory=Header)
    poses: types.sequence[Pose] = field(default_factory=list)


@dataclass
class PoseStamped(IdlStruct, typename="geometry_msgs/msg/PoseStamped"):
    header: Header = field(default_factory=Header)
    pose: Pose = field(default_factory=Pose)


@dataclass
class PoseWithCovariance(IdlStruct, typename="geometry_msgs/msg/PoseWithCovariance"):
    pose: Pose = field(default_factory=Pose)
    covariance: types.array[types.float64, 36] = field(
        default_factory=lambda: np.zeros(36)
    )


@dataclass
class TransformStamped(IdlStruct, typename="geometry_msgs/msg/TransformStamped"):
    header: Header = field(default_factory=Header)
    child_frame_id: str = ""
    transform: Transform = field(default_factory=Transform)


@dataclass
class TwistStamped(IdlStruct, typename="geometry_msgs/msg/TwistStamped"):
    header: Header = field(default_factory=Header)
    twist: Twist = field(default_factory=Twist)


@dataclass
class TwistWithCovariance(IdlStruct, typename="geometry_msgs/msg/TwistWithCovariance"):
    twist: Twist = field(default_factory=Twist)
    covariance: types.array[types.float64, 36] = field(
        default_factory=lambda: np.zeros(36)
    )


@dataclass
class VelocityStamped(IdlStruct, typename="geometry_msgs/msg/VelocityStamped"):
    header: Header = field(default_factory=Header)
    body_frame_id: str = ""
    reference_frame_id: str = ""
    velocity: Twist = field(default_factory=Twist)


@dataclass
class WrenchStamped(IdlStruct, typename="geometry_msgs/msg/WrenchStamped"):
    header: Header = field(default_factory=Header)
    wrench: Wrench = field(default_factory=Wrench)


@dataclass
class AccelWithCovarianceStamped(
    IdlStruct, typename="geometry_msgs/msg/AccelWithCovarianceStamped"
):
    header: Header = field(default_factory=Header)
    accel: AccelWithCovariance = field(default_factory=AccelWithCovariance)


@dataclass
class PolygonInstanceStamped(
    IdlStruct, typename="geometry_msgs/msg/PolygonInstanceStamped"
):
    header: Header = field(default_factory=Header)
    polygon: PolygonInstance = field(default_factory=PolygonInstance)


@dataclass
class PoseWithCovarianceStamped(
    IdlStruct, typename="geometry_msgs/msg/PoseWithCovarianceStamped"
):
    header: Header = field(default_factory=Header)
    pose: PoseWithCovariance = field(default_factory=PoseWithCovariance)


@dataclass
class TwistWithCovarianceStamped(
    IdlStruct, typename="geometry_msgs/msg/TwistWithCovarianceStamped"
):
    header: Header = field(default_factory=Header)
    twist: TwistWithCovariance = field(default_factory=TwistWithCovariance)



from dataclasses import dataclass, field

from ..builtin_interfaces.msg import Time
from ..idl import IdlStruct


@dataclass
class Clock(IdlStruct, typename="rosgraph_msgs/msg/Clock"):
    clock: Time = field(default_factory=Time)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from ..builtin_interfaces.msg import Time
from ..idl import IdlStruct, types


@dataclass
class StatisticDataType(IdlStruct, typename="statistics_msgs/msg/StatisticDataType"):
    STATISTICS_DATA_TYPE_UNINITIALIZED: ClassVar[Literal[0]] = 0
    STATISTICS_DATA_TYPE_AVERAGE: ClassVar[Literal[1]] = 1
    STATISTICS_DATA_TYPE_MINIMUM: ClassVar[Literal[2]] = 2
    STATISTICS_DATA_TYPE_MAXIMUM: ClassVar[Literal[3]] = 3
    STATISTICS_DATA_TYPE_STDDEV: ClassVar[Literal[4]] = 4
    STATISTICS_DATA_TYPE_SAMPLE_COUNT: ClassVar[Literal[5]] = 5


@dataclass
class StatisticDataPoint(IdlStruct, typename="statistics_msgs/msg/StatisticDataPoint"):
    data_type: types.uint8 = 0
    data: types.float64 = 0.0


@dataclass
class MetricsMessage(IdlStruct, typename="statistics_msgs/msg/MetricsMessage"):
    measurement_source_name: str = ""
    metrics_source: str = ""
    unit: str = ""
    window_start: Time = field(default_factory=Time)
    window_stop: Time = field(default_factory=Time)
    statistics: types.sequence[StatisticDataPoint] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from .. import idl
from ..geometry_msgs.msg import PoseStamped, PoseWithCovarianceStamped
from ..nav_msgs.msg import OccupancyGrid, Path
from ..service_msgs.msg import ServiceEventInfo

@dataclass
class GetMap_Request(idl.IdlStruct, typename="nav_msgs/srv/GetMap_Request"):
    pass


@dataclass
class GetMap_Response(idl.IdlStruct, typename="nav_msgs/srv/GetMap_Response"):
    map: OccupancyGrid = field(default_factory=OccupancyGrid)


@dataclass
class GetMap_Event(
    idl.IdlStruct,
    typename="nav_msgs/srv/GetMap_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[GetMap_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[GetMap_Response, 1] = field(default_factory=list)


GetMap: idl.IdlServiceType[GetMap_Request, GetMap_Response, GetMap_Event] = (
    idl.make_idl_service(
        GetMap_Request,
        GetMap_Response,
        event_type=GetMap_Event,
    )
)


@dataclass
class GetPlan_Request(idl.IdlStruct, typename="nav_msgs/srv/GetPlan_Request"):
    start: PoseStamped = field(default_factory=PoseStamped)
    goal: PoseStamped = field(default_factory=PoseStamped)
    tolerance: idl.types.float32 = 0.0


@dataclass
class GetPlan_Response(idl.IdlStruct, typename="nav_msgs/srv/GetPlan_Response"):
    plan: Path = field(default_factory=Path)


@dataclass
class GetPlan_Event(
    idl.IdlStruct,
    typename="nav_msgs/srv/GetPlan_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[GetPlan_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[GetPlan_Response, 1] = field(default_factory=list)


GetPlan: idl.IdlServiceType[GetPlan_Request, GetPlan_Response, GetPlan_Event] = (
    idl.make_idl_service(
        GetPlan_Request,
        GetPlan_Response,
        event_type=GetPlan_Event,
    )
)


@dataclass
class LoadMap_Request(idl.IdlStruct, typename="nav_msgs/srv/LoadMap_Request"):
    map_url: str = ""


@dataclass
class LoadMap_Response(idl.IdlStruct, typename="nav_msgs/srv/LoadMap_Response"):
    RESULT_SUCCESS: ClassVar[Literal[0]] = 0
    RESULT_MAP_DOES_NOT_EXIST: ClassVar[Literal[1]] = 1
    RESULT_INVALID_MAP_DATA: ClassVar[Literal[2]] = 2
    RESULT_INVALID_MAP_METADATA: ClassVar[Literal[3]] = 3
    RESULT_UNDEFINED_FAILURE: ClassVar[Literal[255]] = 255
    map: OccupancyGrid = field(default_factory=OccupancyGrid)
    result: idl.types.uint8 = 0


@dataclass
class LoadMap_Event(
    idl.IdlStruct,
    typename="nav_msgs/srv/LoadMap_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[LoadMap_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[LoadMap_Response, 1] = field(default_factory=list)


LoadMap: idl.IdlServiceType[LoadMap_Request, LoadMap_Response, LoadMap_Event] = (
    idl.make_idl_service(
        LoadMap_Request,
        LoadMap_Response,
        event_type=LoadMap_Event,
    )
)


@dataclass
class SetMap_Request(idl.IdlStruct, typename="nav_msgs/srv/SetMap_Request"):
    map: OccupancyGrid = field(default_factory=OccupancyGrid)
    initial_pose: PoseWithCovarianceStamped = field(
        default_factory=PoseWithCovarianceStamped
    )


@dataclass
class SetMap_Response(idl.IdlStruct, typename="nav_msgs/srv/SetMap_Response"):
    success: bool = False


@dataclass
class SetMap_Event(
    idl.IdlStruct,
    typename="nav_msgs/srv/SetMap_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[SetMap_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[SetMap_Response, 1] = field(default_factory=list)


SetMap: idl.IdlServiceType[SetMap_Request, SetMap_Response, SetMap_Event] = (
    idl.make_idl_service(
        SetMap_Request,
        SetMap_Response,
        event_type=SetMap_Event,
    )
)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from ..builtin_interfaces.msg import Duration
from ..geometry_msgs.msg import Point, Pose, Quaternion, Vector3
from ..idl import IdlStruct, types
from ..sensor_msgs.msg import CompressedImage
from ..std_msgs.msg import ColorRGBA, Header


@dataclass
class ImageMarker(IdlStruct, typename="visualization_msgs/msg/ImageMarker"):
    CIRCLE: ClassVar[Literal[0]] = 0
    LINE_STRIP: ClassVar[Literal[1]] = 1
    LINE_LIST: ClassVar[Literal[2]] = 2
    POLYGON: ClassVar[Literal[3]] = 3
    POINTS: ClassVar[Literal[4]] = 4
    ADD: ClassVar[Literal[0]] = 0
    REMOVE: ClassVar[Literal[1]] = 1
    header: Header = field(default_factory=Header)
    ns: str = ""
    id: types.int32 = 0
    type: types.int32 = 0
    action: types.int32 = 0
    position: Point = field(default_factory=Point)
    scale: types.float32 = 0.0
    outline_color: ColorRGBA = field(default_factory=ColorRGBA)
    filled: types.uint8 = 0
    fill_color: ColorRGBA = field(default_factory=ColorRGBA)
    lifetime: Duration = field(default_factory=Duration)
    points: types.sequence[Point] = field(default_factory=list)
    outline_colors: types.sequence[ColorRGBA] = field(default_factory=list)


@dataclass
class InteractiveMarkerFeedback(
    IdlStruct, typename="visualization_msgs/msg/InteractiveMarkerFeedback"
):
    KEEP_ALIVE: ClassVar[Literal[0]] = 0
    POSE_UPDATE: ClassVar[Literal[1]] = 1
    MENU_SELECT: ClassVar[Literal[2]] = 2
    BUTTON_CLICK: ClassVar[Literal[3]] = 3
    MOUSE_DOWN: ClassVar[Literal[4]] = 4
    MOUSE_UP: ClassVar[Literal[5]] = 5
    header: Header = field(default_factory=Header)
    client_id: str = ""
    marker_name: str = ""
    control_name: str = ""
    event_type: types.uint8 = 0
    pose: Pose = field(default_factory=Pose)
    menu_entry_id: types.uint32 = 0
    mouse_point: Point = field(default_factory=Point)
    mouse_point_valid: bool = False


@dataclass
class InteractiveMarkerPose(
    IdlStruct, typename="visualization_msgs/msg/InteractiveMarkerPose"
):
    header: Header = field(default_factory=Header)
    pose: Pose = field(default_factory=Pose)
    name: str = ""


@dataclass
class MenuEntry(IdlStruct, typename="visualization_msgs/msg/MenuEntry"):
    FEEDBACK: ClassVar[Literal[0]] = 0
    ROSRUN: ClassVar[Literal[1]] = 1
    ROSLAUNCH: ClassVar[Literal[2]] = 2
    id: types.uint32 = 0
    parent_id: types.uint32 = 0
    title: str = ""
    command: str = ""
    command_type: types.uint8 = 0


@dataclass
class MeshFile(IdlStruct, typename="visualization_msgs/msg/MeshFile"):
    filename: str = ""
    data: types.sequence[types.uint8] = field(default_factory=list)


@dataclass
class UVCoordinate(IdlStruct, typename="visualization_msgs/msg/UVCoordinate"):
    u: types.float32 = 0.0
    v: types.float32 = 0.0


@dataclass
class Marker(IdlStruct, typename="visualization_msgs/msg/Marker"):
    ARROW: ClassVar[Literal[0]] = 0
    CUBE: ClassVar[Literal[1]] = 1
    SPHERE: ClassVar[Literal[2]] = 2
    CYLINDER: ClassVar[Literal[3]] = 3
    LINE_STRIP: ClassVar[Literal[4]] = 4
    LINE_LIST: ClassVar[Literal[5]] = 5
    CUBE_LIST: ClassVar[Literal[6]] = 6
    SPHERE_LIST: ClassVar[Literal[7]] = 7
    POINTS: ClassVar[Literal[8]] = 8
    TEXT_VIEW_FACING: ClassVar[Literal[9]] = 9
    MESH_RESOURCE: ClassVar[Literal[10]] = 10
    TRIANGLE_LIST: ClassVar[Literal[11]] = 11
    ARROW_STRIP: ClassVar[Literal[12]] = 12
    ADD: ClassVar[Literal[0]] = 0
    MODIFY: ClassVar[Literal[0]] = 0
    DELETE: ClassVar[Literal[2]] = 2
    DELETEALL: ClassVar[Literal[3]] = 3
    header: Header = field(default_factory=Header)
    ns: str = ""
    id: types.int32 = 0
    type: types.int32 = 0
    action: types.int32 = 0
    pose: Pose = field(default_factory=Pose)
    scale: Vector3 = field(default_factory=Vector3)
    color: ColorRGBA = field(default_factory=ColorRGBA)
    lifetime: Duration = field(default_factory=Duration)
    frame_locked: bool = False
    points: types.sequence[Point] = field(default_factory=list)
    colors: types.sequence[ColorRGBA] = field(default_factory=list)
    texture_resource: str = ""
    texture: CompressedImage = field(default_factory=CompressedImage)
    uv_coordinates: types.sequence[UVCoordinate] = field(default_factory=list)
    text: str = ""
    mesh_resource: str = ""
    mesh_file: MeshFile = field(default_factory=MeshFile)
    mesh_use_embedded_materials: bool = False


@dataclass
class InteractiveMarkerControl(
    IdlStruct, typename="visualization_msgs/msg/InteractiveMarkerControl"
):
    INHERIT: ClassVar[Literal[0]] = 0
    FIXED: ClassVar[Literal[1]] = 1
    VIEW_FACING: ClassVar[Literal[2]] = 2
    NONE: ClassVar[Literal[0]] = 0
    MENU: ClassVar[Literal[1]] = 1
    BUTTON: ClassVar[Literal[2]] = 2
    MOVE_AXIS: ClassVar[Literal[3]] = 3
    MOVE_PLANE: ClassVar[Literal[4]] = 4
    ROTATE_AXIS: ClassVar[Literal[5]] = 5
    MOVE_ROTATE: ClassVar[Literal[6]] = 6
    MOVE_3D: ClassVar[Literal[7]] = 7
    ROTATE_3D: ClassVar[Literal[8]] = 8
    MOVE_ROTATE_3D: ClassVar[Literal[9]] = 9
    name: str = ""
    orientation: Quaternion = field(default_factory=Quaternion)
    orientation_mode: types.uint8 = 0
    interaction_mode: types.uint8 = 0
    always_visible: bool = False
    markers: types.sequence[Marker] = field(default_factory=list)
    independent_marker_orientation: bool = False
    description: str = ""


@dataclass
class MarkerArray(IdlStruct, typename="visualization_msgs/msg/MarkerArray"):
    markers: types.sequence[Marker] = field(default_factory=list)


@dataclass
class InteractiveMarker(IdlStruct, typename="visualization_msgs/msg/InteractiveMarker"):
    header: Header = field(default_factory=Header)
    pose: Pose = field(default_factory=Pose)
    name: str = ""
    description: str = ""
    scale: types.float32 = 0.0
    menu_entries: types.sequence[MenuEntry] = field(default_factory=list)
    controls: types.sequence[InteractiveMarkerControl] = field(default_factory=list)


@dataclass
class InteractiveMarkerInit(
    IdlStruct, typename="visualization_msgs/msg/InteractiveMarkerInit"
):
    server_id: str = ""
    seq_num: types.uint64 = 0
    markers: types.sequence[InteractiveMarker] = field(default_factory=list)


@dataclass
class InteractiveMarkerUpdate(
    IdlStruct, typename="visualization_msgs/msg/InteractiveMarkerUpdate"
):
    KEEP_ALIVE: ClassVar[Literal[0]] = 0
    UPDATE: ClassVar[Literal[1]] = 1
    server_id: str = ""
    seq_num: types.uint64 = 0
    type: types.uint8 = 0
    markers: types.sequence[InteractiveMarker] = field(default_factory=list)
    poses: types.sequence[InteractiveMarkerPose] = field(default_factory=list)
    erases: types.sequence[str] = field(default_factory=list)



from dataclasses import dataclass, field

from .. import idl
from ..service_msgs.msg import ServiceEventInfo
from ..visualization_msgs.msg import InteractiveMarker

@dataclass
class GetInteractiveMarkers_Request(
    idl.IdlStruct, typename="visualization_msgs/srv/GetInteractiveMarkers_Request"
):
    pass


@dataclass
class GetInteractiveMarkers_Response(
    idl.IdlStruct, typename="visualization_msgs/srv/GetInteractiveMarkers_Response"
):
    sequence_number: idl.types.uint64 = 0
    markers: idl.types.sequence[InteractiveMarker] = field(default_factory=list)


@dataclass
class GetInteractiveMarkers_Event(
    idl.IdlStruct,
    typename="visualization_msgs/srv/GetInteractiveMarkers_Event"
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[GetInteractiveMarkers_Request, 1] = field(
        default_factory=list
    )
    response: idl.types.sequence[GetInteractiveMarkers_Response, 1] = field(
        default_factory=list
    )


GetInteractiveMarkers: idl.IdlServiceType[
    GetInteractiveMarkers_Request,
    GetInteractiveMarkers_Response,
    GetInteractiveMarkers_Event,
] = idl.make_idl_service(
    GetInteractiveMarkers_Request,
    GetInteractiveMarkers_Response,
    event_type=GetInteractiveMarkers_Event,
)



from dataclasses import dataclass
from ..idl import IdlStruct, Time, types

Time.__module__ = __name__

@dataclass
class Duration(IdlStruct, typename="builtin_interfaces/msg/Duration"):
    sec: types.int32 = 0
    nanosec: types.uint32 = 0



from dataclasses import dataclass, field

from .. import idl
from ..diagnostic_msgs.msg import DiagnosticStatus
from ..service_msgs.msg import ServiceEventInfo

@dataclass
class AddDiagnostics_Request(
    idl.IdlStruct, typename="diagnostic_msgs/srv/AddDiagnostics_Request"
):
    load_namespace: str = ""


@dataclass
class AddDiagnostics_Response(
    idl.IdlStruct, typename="diagnostic_msgs/srv/AddDiagnostics_Response"
):
    success: bool = False
    message: str = ""


@dataclass
class AddDiagnostics_Event(
    idl.IdlStruct,
    typename="diagnostic_msgs/srv/AddDiagnostics_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[AddDiagnostics_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[AddDiagnostics_Response, 1] = field(default_factory=list)


AddDiagnostics: idl.IdlServiceType[
    AddDiagnostics_Request,
    AddDiagnostics_Response,
    AddDiagnostics_Event,
] = idl.make_idl_service(
    AddDiagnostics_Request,
    AddDiagnostics_Response,
    event_type=AddDiagnostics_Event,
)


@dataclass
class SelfTest_Request(idl.IdlStruct, typename="diagnostic_msgs/srv/SelfTest_Request"):
    pass


@dataclass
class SelfTest_Response(idl.IdlStruct, typename="diagnostic_msgs/srv/SelfTest_Response"):
    id: str = ""
    passed: idl.types.byte = 0
    status: idl.types.sequence[DiagnosticStatus] = field(default_factory=list)


@dataclass
class SelfTest_Event(
    idl.IdlStruct,
    typename="diagnostic_msgs/srv/SelfTest_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[SelfTest_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[SelfTest_Response, 1] = field(default_factory=list)


SelfTest: idl.IdlServiceType[SelfTest_Request, SelfTest_Response, SelfTest_Event] = (
    idl.make_idl_service(
        SelfTest_Request,
        SelfTest_Response,
        event_type=SelfTest_Event,
    )
)



import array
import asyncio
import inspect
from dataclasses import fields, is_dataclass
from typing import List, Mapping, Sequence, Tuple, Type

import asyncio_for_robotics.ros2 as afor
import numpy as np
import pytest
import rclpy
from rclpy.subscription import Subscription
from utils import assert_strictly_eq

from ros2_pyterfaces import all_msgs, idl
from ros2_pyterfaces.all_msgs import (
    Char,
    DiagnosticStatus,
    Empty,
    Float32,
    Float64,
    Header,
    Imu,
    JointState,
    KeyValue,
    Quaternion,
    String,
    Time,
    Trajectory,
    TrajectoryPoint,
)

TYPES: List[Type[idl.IdlStruct]] = sorted(
    {
        obj
        for obj in vars(all_msgs).values()
        if inspect.isclass(obj)
        and issubclass(obj, idl.IdlStruct)
        and obj is not idl.IdlStruct
    },
    key=lambda cls: cls.__name__,
)
NOT_IN_ROS = [
    # not in ROS yet
    "nav_msgs/msg/Trajectory",
    "nav_msgs/msg/TrajectoryPoint",
]
TYPES = [t for t in TYPES if t.get_type_name() not in NOT_IN_ROS]
# TYPES = [Empty, JointState]


class RawSub(afor.Sub):
    def _resolve_sub(self, topic_info: afor.TopicInfo) -> Subscription:
        with self.session.lock() as node:
            return node.create_subscription(
                **topic_info.as_kwarg(), callback=self.callback_for_sub, raw=True
            )


@pytest.fixture(scope="module")
def init():
    rclpy.init()
    ses = afor.auto_session()
    yield
    ses.close()
    try:
        rclpy.shutdown()
    except:
        pass


async def send_raw(t: Type[idl.IdlStruct] | idl.IdlStruct, ros_t):
    tinfo = afor.TopicInfo(t.get_type_name(), ros_t)
    sub = afor.Sub(**tinfo.as_kwarg())

    async def periodic_pub():
        with afor.auto_session().lock() as node:
            pub = node.create_publisher(**tinfo.as_kwarg())
            if isinstance(t, idl.IdlStruct):
                pub.publish(t.serialize())
            else:
                pub.publish(t().serialize())

    try:
        async with asyncio.TaskGroup() as tg:
            tg.create_task(periodic_pub())
            recv = await sub.wait_for_value()
            res = t.from_ros(recv)
            return res
    finally:
        sub.close()


async def recv_raw(t: Type[idl.IdlStruct], ros_t):
    tinfo = afor.TopicInfo(t.get_type_name(), ros_t)
    sub = RawSub(**tinfo.as_kwarg())

    async def periodic_pub():
        with afor.auto_session().lock() as node:
            pub = node.create_publisher(**tinfo.as_kwarg())
            pub.publish(ros_t())

    try:
        async with asyncio.TaskGroup() as tg:
            tg.create_task(periodic_pub())
            recv = await sub.wait_for_value()
            res = t.deserialize(recv)
            return res
    finally:
        sub.close()


@pytest.mark.parametrize("my_type", TYPES)
async def test_send_raw(my_type: Type[idl.IdlStruct], init):
    res = await afor.soft_wait_for(send_raw(my_type, my_type.get_ros_type()), 3)
    assert not isinstance(res, TimeoutError)
    # assert res == my_type()
    assert_strictly_eq(res, my_type())


@pytest.mark.parametrize("my_type", TYPES)
async def test_recv_raw(my_type: Type[idl.IdlStruct], init):
    res = await afor.soft_wait_for(recv_raw(my_type, my_type.get_ros_type()), 3)
    assert not isinstance(res, TimeoutError)
    # assert res == my_type()
    assert_strictly_eq(res, my_type())


VALUES = [
    String(data="hello"),
    Char(3),
    Float32(np.float32(np.pi)),
    Float64(np.pi),
    JointState(
        header=Header(stamp=Time(1, 76)),
        position=list(range(100)),
        velocity=list(range(100)),
    ),
    Empty(),
    Imu(orientation=Quaternion(1, 0, 0, 0), orientation_covariance=np.full((9,), 0.1)),
    DiagnosticStatus(
        level=DiagnosticStatus.WARN,
        name="hey",
        message="hello",
        values=[KeyValue(key="heyyo", value="yey")],
    ),
]


@pytest.mark.parametrize("my_msg", VALUES)
async def test_send_raw_values(my_msg: idl.IdlStruct, init):
    res = await afor.soft_wait_for(send_raw(my_msg, my_msg.get_ros_type()), 3)
    assert not isinstance(res, TimeoutError)
    # assert res == my_type()
    assert_strictly_eq(res, my_msg)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from ..idl import IdlStruct, types


@dataclass
class FieldType(IdlStruct, typename="type_description_interfaces/msg/FieldType"):
    FIELD_TYPE_NOT_SET: ClassVar[Literal[0]] = 0
    FIELD_TYPE_NESTED_TYPE: ClassVar[Literal[1]] = 1
    FIELD_TYPE_INT8: ClassVar[Literal[2]] = 2
    FIELD_TYPE_UINT8: ClassVar[Literal[3]] = 3
    FIELD_TYPE_INT16: ClassVar[Literal[4]] = 4
    FIELD_TYPE_UINT16: ClassVar[Literal[5]] = 5
    FIELD_TYPE_INT32: ClassVar[Literal[6]] = 6
    FIELD_TYPE_UINT32: ClassVar[Literal[7]] = 7
    FIELD_TYPE_INT64: ClassVar[Literal[8]] = 8
    FIELD_TYPE_UINT64: ClassVar[Literal[9]] = 9
    FIELD_TYPE_FLOAT: ClassVar[Literal[10]] = 10
    FIELD_TYPE_DOUBLE: ClassVar[Literal[11]] = 11
    FIELD_TYPE_LONG_DOUBLE: ClassVar[Literal[12]] = 12
    FIELD_TYPE_CHAR: ClassVar[Literal[13]] = 13
    FIELD_TYPE_WCHAR: ClassVar[Literal[14]] = 14
    FIELD_TYPE_BOOLEAN: ClassVar[Literal[15]] = 15
    FIELD_TYPE_BYTE: ClassVar[Literal[16]] = 16
    FIELD_TYPE_STRING: ClassVar[Literal[17]] = 17
    FIELD_TYPE_WSTRING: ClassVar[Literal[18]] = 18
    FIELD_TYPE_FIXED_STRING: ClassVar[Literal[19]] = 19
    FIELD_TYPE_FIXED_WSTRING: ClassVar[Literal[20]] = 20
    FIELD_TYPE_BOUNDED_STRING: ClassVar[Literal[21]] = 21
    FIELD_TYPE_BOUNDED_WSTRING: ClassVar[Literal[22]] = 22
    FIELD_TYPE_NESTED_TYPE_ARRAY: ClassVar[Literal[49]] = 49
    FIELD_TYPE_INT8_ARRAY: ClassVar[Literal[50]] = 50
    FIELD_TYPE_UINT8_ARRAY: ClassVar[Literal[51]] = 51
    FIELD_TYPE_INT16_ARRAY: ClassVar[Literal[52]] = 52
    FIELD_TYPE_UINT16_ARRAY: ClassVar[Literal[53]] = 53
    FIELD_TYPE_INT32_ARRAY: ClassVar[Literal[54]] = 54
    FIELD_TYPE_UINT32_ARRAY: ClassVar[Literal[55]] = 55
    FIELD_TYPE_INT64_ARRAY: ClassVar[Literal[56]] = 56
    FIELD_TYPE_UINT64_ARRAY: ClassVar[Literal[57]] = 57
    FIELD_TYPE_FLOAT_ARRAY: ClassVar[Literal[58]] = 58
    FIELD_TYPE_DOUBLE_ARRAY: ClassVar[Literal[59]] = 59
    FIELD_TYPE_LONG_DOUBLE_ARRAY: ClassVar[Literal[60]] = 60
    FIELD_TYPE_CHAR_ARRAY: ClassVar[Literal[61]] = 61
    FIELD_TYPE_WCHAR_ARRAY: ClassVar[Literal[62]] = 62
    FIELD_TYPE_BOOLEAN_ARRAY: ClassVar[Literal[63]] = 63
    FIELD_TYPE_BYTE_ARRAY: ClassVar[Literal[64]] = 64
    FIELD_TYPE_STRING_ARRAY: ClassVar[Literal[65]] = 65
    FIELD_TYPE_WSTRING_ARRAY: ClassVar[Literal[66]] = 66
    FIELD_TYPE_FIXED_STRING_ARRAY: ClassVar[Literal[67]] = 67
    FIELD_TYPE_FIXED_WSTRING_ARRAY: ClassVar[Literal[68]] = 68
    FIELD_TYPE_BOUNDED_STRING_ARRAY: ClassVar[Literal[69]] = 69
    FIELD_TYPE_BOUNDED_WSTRING_ARRAY: ClassVar[Literal[70]] = 70
    FIELD_TYPE_NESTED_TYPE_BOUNDED_SEQUENCE: ClassVar[Literal[97]] = 97
    FIELD_TYPE_INT8_BOUNDED_SEQUENCE: ClassVar[Literal[98]] = 98
    FIELD_TYPE_UINT8_BOUNDED_SEQUENCE: ClassVar[Literal[99]] = 99
    FIELD_TYPE_INT16_BOUNDED_SEQUENCE: ClassVar[Literal[100]] = 100
    FIELD_TYPE_UINT16_BOUNDED_SEQUENCE: ClassVar[Literal[101]] = 101
    FIELD_TYPE_INT32_BOUNDED_SEQUENCE: ClassVar[Literal[102]] = 102
    FIELD_TYPE_UINT32_BOUNDED_SEQUENCE: ClassVar[Literal[103]] = 103
    FIELD_TYPE_INT64_BOUNDED_SEQUENCE: ClassVar[Literal[104]] = 104
    FIELD_TYPE_UINT64_BOUNDED_SEQUENCE: ClassVar[Literal[105]] = 105
    FIELD_TYPE_FLOAT_BOUNDED_SEQUENCE: ClassVar[Literal[106]] = 106
    FIELD_TYPE_DOUBLE_BOUNDED_SEQUENCE: ClassVar[Literal[107]] = 107
    FIELD_TYPE_LONG_DOUBLE_BOUNDED_SEQUENCE: ClassVar[Literal[108]] = 108
    FIELD_TYPE_CHAR_BOUNDED_SEQUENCE: ClassVar[Literal[109]] = 109
    FIELD_TYPE_WCHAR_BOUNDED_SEQUENCE: ClassVar[Literal[110]] = 110
    FIELD_TYPE_BOOLEAN_BOUNDED_SEQUENCE: ClassVar[Literal[111]] = 111
    FIELD_TYPE_BYTE_BOUNDED_SEQUENCE: ClassVar[Literal[112]] = 112
    FIELD_TYPE_STRING_BOUNDED_SEQUENCE: ClassVar[Literal[113]] = 113
    FIELD_TYPE_WSTRING_BOUNDED_SEQUENCE: ClassVar[Literal[114]] = 114
    FIELD_TYPE_FIXED_STRING_BOUNDED_SEQUENCE: ClassVar[Literal[115]] = 115
    FIELD_TYPE_FIXED_WSTRING_BOUNDED_SEQUENCE: ClassVar[Literal[116]] = 116
    FIELD_TYPE_BOUNDED_STRING_BOUNDED_SEQUENCE: ClassVar[Literal[117]] = 117
    FIELD_TYPE_BOUNDED_WSTRING_BOUNDED_SEQUENCE: ClassVar[Literal[118]] = 118
    FIELD_TYPE_NESTED_TYPE_UNBOUNDED_SEQUENCE: ClassVar[Literal[145]] = 145
    FIELD_TYPE_INT8_UNBOUNDED_SEQUENCE: ClassVar[Literal[146]] = 146
    FIELD_TYPE_UINT8_UNBOUNDED_SEQUENCE: ClassVar[Literal[147]] = 147
    FIELD_TYPE_INT16_UNBOUNDED_SEQUENCE: ClassVar[Literal[148]] = 148
    FIELD_TYPE_UINT16_UNBOUNDED_SEQUENCE: ClassVar[Literal[149]] = 149
    FIELD_TYPE_INT32_UNBOUNDED_SEQUENCE: ClassVar[Literal[150]] = 150
    FIELD_TYPE_UINT32_UNBOUNDED_SEQUENCE: ClassVar[Literal[151]] = 151
    FIELD_TYPE_INT64_UNBOUNDED_SEQUENCE: ClassVar[Literal[152]] = 152
    FIELD_TYPE_UINT64_UNBOUNDED_SEQUENCE: ClassVar[Literal[153]] = 153
    FIELD_TYPE_FLOAT_UNBOUNDED_SEQUENCE: ClassVar[Literal[154]] = 154
    FIELD_TYPE_DOUBLE_UNBOUNDED_SEQUENCE: ClassVar[Literal[155]] = 155
    FIELD_TYPE_LONG_DOUBLE_UNBOUNDED_SEQUENCE: ClassVar[Literal[156]] = 156
    FIELD_TYPE_CHAR_UNBOUNDED_SEQUENCE: ClassVar[Literal[157]] = 157
    FIELD_TYPE_WCHAR_UNBOUNDED_SEQUENCE: ClassVar[Literal[158]] = 158
    FIELD_TYPE_BOOLEAN_UNBOUNDED_SEQUENCE: ClassVar[Literal[159]] = 159
    FIELD_TYPE_BYTE_UNBOUNDED_SEQUENCE: ClassVar[Literal[160]] = 160
    FIELD_TYPE_STRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[161]] = 161
    FIELD_TYPE_WSTRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[162]] = 162
    FIELD_TYPE_FIXED_STRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[163]] = 163
    FIELD_TYPE_FIXED_WSTRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[164]] = 164
    FIELD_TYPE_BOUNDED_STRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[165]] = 165
    FIELD_TYPE_BOUNDED_WSTRING_UNBOUNDED_SEQUENCE: ClassVar[Literal[166]] = 166

    type_id: types.uint8 = 0
    capacity: types.uint64 = 0
    string_capacity: types.uint64 = 0
    nested_type_name: types.bounded_str[255] = ""


@dataclass
class Field(IdlStruct, typename="type_description_interfaces/msg/Field"):
    name: str = ""
    type: FieldType = field(default_factory=FieldType)
    default_value: str = ""


@dataclass
class IndividualTypeDescription(
    IdlStruct, typename="type_description_interfaces/msg/IndividualTypeDescription"
):
    type_name: types.bounded_str[255] = ""
    fields: types.sequence[Field] = field(default_factory=list)


@dataclass
class KeyValue(IdlStruct, typename="type_description_interfaces/msg/KeyValue"):
    key: str = ""
    value: str = ""


@dataclass
class TypeDescription(
    IdlStruct, typename="type_description_interfaces/msg/TypeDescription"
):
    type_description: IndividualTypeDescription = field(
        default_factory=IndividualTypeDescription
    )
    referenced_type_descriptions: types.sequence[IndividualTypeDescription] = field(
        default_factory=list
    )


@dataclass
class TypeSource(IdlStruct, typename="type_description_interfaces/msg/TypeSource"):
    type_name: str = ""
    encoding: str = ""
    raw_file_contents: str = ""




from dataclasses import dataclass, field

from ..builtin_interfaces.msg import Duration, Time
from ..idl import IdlStruct


@dataclass
class Builtins(IdlStruct, typename="test_msgs/msg/Builtins"):
    duration_value: Duration = field(default_factory=Duration)
    time_value: Time = field(default_factory=Time)



from dataclasses import dataclass, field
from ..idl import IdlStruct, types
from ..sensor_msgs.msg import Image, RegionOfInterest
from ..std_msgs.msg import Header

@dataclass
class DisparityImage(IdlStruct, typename="stereo_msgs/msg/DisparityImage"):
    header: Header = field(default_factory=Header)
    image: Image = field(default_factory=Image)
    f: types.float32 = 0.0
    t: types.float32 = 0.0
    valid_window: RegionOfInterest = field(default_factory=RegionOfInterest)
    min_disparity: types.float32 = 0.0
    max_disparity: types.float32 = 0.0
    delta_d: types.float32 = 0.0



import inspect
import json
import subprocess
import time
from pathlib import Path
from pprint import pprint
from typing import Dict, List, Tuple, Type

import asyncio_for_robotics.ros2 as afor
import pytest
import rclpy
import rclpy._rclpy_pybind11 as _impl

from ros2_pyterfaces import all_msgs, idl
from ros2_pyterfaces.all_msgs import (
    DiagnosticStatus,
    JointState,
    KeyValue,
    String,
    Trajectory,
    TrajectoryPoint,
)
from ros2_pyterfaces.std_msgs.msg import Char, Empty, Float32

NOT_IMPL_TYPES = [Trajectory, TrajectoryPoint]
TYPES: List[Type[idl.IdlStruct]] = sorted(
    {
        obj
        for obj in vars(all_msgs).values()
        if inspect.isclass(obj)
        and issubclass(obj, idl.IdlStruct)
        and obj is not idl.IdlStruct
    },
    key=lambda cls: cls.__name__,
)

# TYPES = [
#     Empty,
#     Float32,
#     TrajectoryPoint,
#     Trajectory,
#     Char,
#     DiagnosticStatus,
# ]


def get_name_hash(my_type: Type[idl.IdlStruct]) -> Tuple[str, str]:
    sub = afor.Sub(my_type.get_ros_type(), f"/get_msg_hash/{my_type.get_type_name()}")
    try:
        with afor.auto_session().lock() as node:
            info = _impl.rclpy_get_subscriptions_info_by_topic(
                node.handle, sub.topic_info.topic, False
            )
        # pprint(info)
        return info[0]["topic_type"], info[0]["topic_type_hash"]["value"]
    finally:
        sub.close()


@pytest.fixture(scope="module")
def init():
    rclpy.init()
    ses = afor.auto_session()
    yield
    ses.close()
    try:
        rclpy.shutdown()
    except:
        pass


@pytest.mark.parametrize("my_type", TYPES)
async def test_hashes(my_type: Type[idl.IdlStruct], init):
    try:
        name, hash = get_name_hash(my_type)
    except AttributeError as e:
        if "has no attribute" in str(e) and my_type in NOT_IMPL_TYPES:
            pytest.xfail(
                f"Message {my_type.get_type_name()} doesn't exist in ROS 2",
            )
        raise e

    assert my_type.get_type_name() == name
    assert my_type._hash_rihs01_raw().digest() == hash



from typing import Tuple

import pytest

import ros2_pyterfaces
from ros2_pyterfaces import idl
from ros2_pyterfaces.diagnostic_msgs.msg import KeyValue
from ros2_pyterfaces.sensor_msgs.msg import JointState
from ros2_pyterfaces.std_msgs.msg import String


@pytest.mark.parametrize(
    "couple",
    [
        (
            String,
            "RIHS01_df668c740482bbd48fb39d76a70dfd4bd59db1288021743503259e948f6b1a18",
        ),
        (
            KeyValue,
            "RIHS01_d68081eaa540288c5440753baecef0c4e16e81a5f78ad68902ded5100413bb42",
        ),
        (
            JointState,
            "RIHS01_a13ee3a330e346c9d87b5aa18d24e11690752bd33a0350f11c5882bc9179260e",
        ),
    ],
)
def test_hashes(couple: Tuple[idl.IdlStruct, str]):
    assert couple[0].hash_rihs01() == couple[1]





import inspect
from importlib import import_module
from typing import List, Tuple, Type

import numpy as np
import pytest
from rclpy.serialization import deserialize_message, serialize_message
from utils import assert_strictly_eq

from ros2_pyterfaces import all_msgs, idl
from ros2_pyterfaces.all_msgs import (
    Char,
    DiagnosticStatus,
    Empty,
    Float32,
    Imu,
    JointState,
    KeyValue,
    String,
    Trajectory,
    TrajectoryPoint,
)
from ros2_pyterfaces.builtin_interfaces.msg import Time
from ros2_pyterfaces.geometry_msgs.msg import Quaternion
from ros2_pyterfaces.std_msgs.msg import Float64, Header

TYPES: List[Type[idl.IdlStruct]] = [
    obj
    for obj in vars(all_msgs).values()
    if inspect.isclass(obj)
    and issubclass(obj, idl.IdlStruct)
    and obj is not idl.IdlStruct
]
NOT_IN_ROS = [
    # not in ROS yet
    "nav_msgs/msg/Trajectory",
    "nav_msgs/msg/TrajectoryPoint",
]
TYPES = [t for t in TYPES if t.get_type_name() not in NOT_IN_ROS]
# TYPES = [Empty, KeyValue, JointState, Char, Imu]


@pytest.mark.parametrize("my_type", TYPES)
def test_deserialize(my_type: Type[idl.IdlStruct]):
    # We cannot compare bit be bit the serialization because of padding
    ros_msg_type = my_type.get_ros_type()
    idl_from_ros = my_type.deserialize(serialize_message(ros_msg_type()))
    # easier to compare that way == doesn't work on arrays
    assert_strictly_eq(idl_from_ros, my_type())


@pytest.mark.parametrize("my_type", TYPES)
def test_serialize(my_type: Type[idl.IdlStruct]):
    # We cannot compare bit be bit the serialization because of padding
    ros_msg_type = my_type.get_ros_type()
    ros_from_idl = deserialize_message(my_type().serialize(), ros_msg_type)
    # easier to compare that way == doesn't work on arrays
    assert ros_from_idl == ros_msg_type()


VALUES = [
    String(data="hello"),
    Char(3),
    Float32(np.float32(np.pi)),
    Float64(np.pi),
    JointState(
        header=Header(stamp=Time(1, 76)),
        position=list(range(100)),
        velocity=list(range(100)),
    ),
    Empty(),
    Imu(orientation=Quaternion(1, 0, 0, 0), orientation_covariance=np.full((9,), 0.1)),
    DiagnosticStatus(
        level=DiagnosticStatus.WARN,
        name="hey",
        message="hello",
        values=[KeyValue(key="heyyo", value="yey")],
    ),
]


@pytest.mark.parametrize("my_msg", VALUES)
def test_deserialize_values(my_msg: idl.IdlStruct):
    # We cannot compare bit be bit the serialization because of padding
    ros_msg = my_msg.to_ros()
    idl_from_ros = my_msg.deserialize(serialize_message(ros_msg))
    # assert my_msg == my_msg.from_ros(ros_from_idl)
    assert_strictly_eq(my_msg, idl_from_ros)

@pytest.mark.parametrize("my_msg", VALUES)
def test_serialize_values(my_msg: idl.IdlStruct):
    # We cannot compare bit be bit the serialization because of padding
    ros_msg_type = my_msg.get_ros_type()
    ros_from_idl = deserialize_message(my_msg.serialize(), ros_msg_type)
    # assert my_msg == my_msg.from_ros(ros_from_idl)
    assert_strictly_eq(my_msg, my_msg.from_ros(ros_from_idl))

    back_to_idl = my_msg.deserialize(serialize_message(ros_from_idl))
    # easier to compare that way == doesn't work on arrays
    assert_strictly_eq(my_msg, back_to_idl)



from .builtin_interfaces.msg import *
from .diagnostic_msgs.msg import KeyValue as DiagnosticMsgs_KeyValue
from .diagnostic_msgs.msg import *
from .geometry_msgs.msg import *
from .lifecycle_msgs.msg import *
from .nav_msgs.msg import *
from .rcl_interfaces.msg import *
from .rosgraph_msgs.msg import *
from .sensor_msgs.msg import *
from .service_msgs.msg import *
from .shape_msgs.msg import *
from .statistics_msgs.msg import *
from .std_msgs.msg import *
from .stereo_msgs.msg import *
from .trajectory_msgs.msg import *
from .type_description_interfaces.msg import Field, FieldType, IndividualTypeDescription
from .type_description_interfaces.msg import (
    KeyValue as TypeDescriptionInterfaces_KeyValue,
)
from .type_description_interfaces.msg import TypeDescription, TypeSource
from .unique_identifier_msgs.msg import *
from .visualization_msgs.msg import *



from dataclasses import dataclass, field
from typing import ClassVar, Literal

from .. import idl
from .msg import (
    ListParametersResult,
    LoggerLevel,
    Parameter,
    ParameterDescriptor,
    ParameterValue,
    SetLoggerLevelsResult,
    SetParametersResult,
)


@dataclass
class DescribeParameters_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/DescribeParameters_Request"
):
    names: idl.types.sequence[str] = field(default_factory=list)


@dataclass
class DescribeParameters_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/DescribeParameters_Response"
):
    descriptors: idl.types.sequence[ParameterDescriptor] = field(default_factory=list)


DescribeParameters = idl.make_idl_service(
    DescribeParameters_Request, DescribeParameters_Response
)
DescribeParameters_Event = DescribeParameters.Event


@dataclass
class GetLoggerLevels_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetLoggerLevels_Request"
):
    names: idl.types.sequence[str] = field(default_factory=list)


@dataclass
class GetLoggerLevels_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetLoggerLevels_Response"
):
    levels: idl.types.sequence[LoggerLevel] = field(default_factory=list)


GetLoggerLevels = idl.make_idl_service(GetLoggerLevels_Request, GetLoggerLevels_Response)
GetLoggerLevels_Event = GetLoggerLevels.Event


@dataclass
class GetParameters_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetParameters_Request"
):
    names: idl.types.sequence[str] = field(default_factory=list)


@dataclass
class GetParameters_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetParameters_Response"
):
    values: idl.types.sequence[ParameterValue] = field(default_factory=list)


GetParameters = idl.make_idl_service(GetParameters_Request, GetParameters_Response)
GetParameters_Event = GetParameters.Event


@dataclass
class GetParameterTypes_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetParameterTypes_Request"
):
    names: idl.types.sequence[str] = field(default_factory=list)


@dataclass
class GetParameterTypes_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/GetParameterTypes_Response"
):
    types: idl.types.sequence[idl.types.uint8] = field(default_factory=list)


GetParameterTypes = idl.make_idl_service(
    GetParameterTypes_Request, GetParameterTypes_Response
)
GetParameterTypes_Event = GetParameterTypes.Event


@dataclass
class ListParameters_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/ListParameters_Request"
):
    DEPTH_RECURSIVE: ClassVar[Literal[0]] = 0

    prefixes: idl.types.sequence[str] = field(default_factory=list)
    depth: idl.types.uint64 = 0


@dataclass
class ListParameters_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/ListParameters_Response"
):
    result: ListParametersResult = field(default_factory=ListParametersResult)


ListParameters = idl.make_idl_service(ListParameters_Request, ListParameters_Response)
ListParameters_Event = ListParameters.Event


@dataclass
class SetLoggerLevels_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetLoggerLevels_Request"
):
    levels: idl.types.sequence[LoggerLevel] = field(default_factory=list)


@dataclass
class SetLoggerLevels_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetLoggerLevels_Response"
):
    results: idl.types.sequence[SetLoggerLevelsResult] = field(default_factory=list)


SetLoggerLevels = idl.make_idl_service(
    SetLoggerLevels_Request, SetLoggerLevels_Response
)
SetLoggerLevels_Event = SetLoggerLevels.Event


@dataclass
class SetParametersAtomically_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetParametersAtomically_Request"
):
    parameters: idl.types.sequence[Parameter] = field(default_factory=list)


@dataclass
class SetParametersAtomically_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetParametersAtomically_Response"
):
    result: SetParametersResult = field(default_factory=SetParametersResult)


SetParametersAtomically = idl.make_idl_service(
    SetParametersAtomically_Request, SetParametersAtomically_Response
)
SetParametersAtomically_Event = SetParametersAtomically.Event


@dataclass
class SetParameters_Request(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetParameters_Request"
):
    parameters: idl.types.sequence[Parameter] = field(default_factory=list)


@dataclass
class SetParameters_Response(
    idl.IdlStruct, typename="rcl_interfaces/srv/SetParameters_Response"
):
    results: idl.types.sequence[SetParametersResult] = field(default_factory=list)


SetParameters = idl.make_idl_service(SetParameters_Request, SetParameters_Response)
SetParameters_Event = SetParameters.Event



from dataclasses import dataclass, field
from ..idl import IdlStruct, types
from ..std_msgs.msg import Header
from ..builtin_interfaces.msg import Duration
from ..geometry_msgs.msg import Transform, Twist

@dataclass
class JointTrajectoryPoint(IdlStruct, typename="trajectory_msgs/msg/JointTrajectoryPoint"):
    positions: types.sequence[types.float64] = field(default_factory=list)
    velocities: types.sequence[types.float64] = field(default_factory=list)
    accelerations: types.sequence[types.float64] = field(default_factory=list)
    effort: types.sequence[types.float64] = field(default_factory=list)
    time_from_start: Duration = field(default_factory=Duration)


@dataclass
class MultiDOFJointTrajectoryPoint(IdlStruct, typename="trajectory_msgs/msg/MultiDOFJointTrajectoryPoint"):
    transforms: types.sequence[Transform] = field(default_factory=list)
    velocities: types.sequence[Twist] = field(default_factory=list)
    accelerations: types.sequence[Twist] = field(default_factory=list)
    time_from_start: Duration = field(default_factory=Duration)


@dataclass
class JointTrajectory(IdlStruct, typename="trajectory_msgs/msg/JointTrajectory"):
    header: Header = field(default_factory=Header)
    joint_names: types.sequence[str] = field(default_factory=list)
    points: types.sequence[JointTrajectoryPoint] = field(default_factory=list)


@dataclass
class MultiDOFJointTrajectory(IdlStruct, typename="trajectory_msgs/msg/MultiDOFJointTrajectory"):
    header: Header = field(default_factory=Header)
    joint_names: types.sequence[str] = field(default_factory=list)
    points: types.sequence[MultiDOFJointTrajectoryPoint] = field(default_factory=list)



import inspect
from typing import List, Type

import numpy as np
import pytest
from utils import assert_msg_equal_as_lists, assert_strictly_eq

from ros2_pyterfaces import all_msgs, idl
from ros2_pyterfaces.all_msgs import DiagnosticStatus, Empty, Float32, KeyValue, String
from ros2_pyterfaces.builtin_interfaces.msg import Time
from ros2_pyterfaces.geometry_msgs.msg import Quaternion, Vector3
from ros2_pyterfaces.sensor_msgs.msg import Imu, JointState
from ros2_pyterfaces.std_msgs.msg import Float64, Header

TYPES: List[Type[idl.IdlStruct]] = [
    obj
    for obj in vars(all_msgs).values()
    if inspect.isclass(obj)
    and issubclass(obj, idl.IdlStruct)
    and obj is not idl.IdlStruct
]
NOT_IN_ROS = [
    "nav_msgs/msg/Trajectory",
    "nav_msgs/msg/TrajectoryPoint",
]
TYPES = [t for t in TYPES if t.get_type_name() not in NOT_IN_ROS]

VALUES = [
    KeyValue(key="mode", value="auto"),
    String(data="hello"),
    Float32(data=np.float32(np.pi)),
    Float64(data=np.pi),
    JointState(
        header=Header(stamp=Time(1, 76)),
        name=["left", "right"],
        position=list(range(5)),
        velocity=[0.1, 0.2, 0.3],
    ),
    Empty(),
    Vector3(1, 2, 3),
    Imu(
        orientation=Quaternion(1, 0, 0, 0),
        orientation_covariance=np.full((9,), 0.1),
    ),
    DiagnosticStatus(
        level=DiagnosticStatus.WARN,
        name="status",
        message="hello",
        values=[KeyValue(key="mode", value="auto")],
    ),
]


@pytest.mark.parametrize("my_type", TYPES)
def test_to_ros_defaults(my_type: Type[idl.IdlStruct]):
    ros_msg_type = my_type.to_ros_type()
    ros_msg = my_type().to_ros()

    assert isinstance(ros_msg, ros_msg_type)
    assert_msg_equal_as_lists(ros_msg, ros_msg_type())


@pytest.mark.parametrize("my_type", TYPES)
def test_from_ros_defaults(my_type: Type[idl.IdlStruct]):
    ros_msg = my_type.to_ros_type()()
    idl_msg = my_type.from_ros(ros_msg)

    assert_strictly_eq(idl_msg, my_type())


@pytest.mark.parametrize("my_msg", VALUES)
def test_to_ros_from_ros_values_roundtrip(my_msg: idl.IdlStruct):
    ros_msg = my_msg.to_ros()

    assert isinstance(ros_msg, type(my_msg).to_ros_type())
    assert_strictly_eq(type(my_msg).from_ros(ros_msg), my_msg)


def test_to_ros_coerces_scalar_float_fields():
    float32_msg = Float32(data=1).to_ros()
    float64_msg = Float64(data=1).to_ros()
    vector3_msg = Vector3(1, 2, 3).to_ros()

    assert isinstance(float32_msg.data, float)
    assert isinstance(float64_msg.data, float)
    assert isinstance(vector3_msg.x, float)
    assert isinstance(vector3_msg.y, float)
    assert isinstance(vector3_msg.z, float)



from dataclasses import dataclass, field

from .. import idl
from ..service_msgs.msg import ServiceEventInfo


@dataclass
class Empty_Request(idl.IdlStruct, typename="std_srvs/srv/Empty_Request"):
    pass


@dataclass
class Empty_Response(idl.IdlStruct, typename="std_srvs/srv/Empty_Response"):
    pass


@dataclass
class Empty_Event(
    idl.IdlStruct,
    typename="std_srvs/srv/Empty_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[Empty_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[Empty_Response, 1] = field(default_factory=list)


Empty: idl.IdlServiceType[Empty_Request, Empty_Response, Empty_Event] = (
    idl.make_idl_service(
        Empty_Request,
        Empty_Response,
        event_type=Empty_Event,
    )
)


@dataclass
class SetBool_Request(idl.IdlStruct, typename="std_srvs/srv/SetBool_Request"):
    data: bool = False


@dataclass
class SetBool_Response(idl.IdlStruct, typename="std_srvs/srv/SetBool_Response"):
    success: bool = False
    message: str = ""


@dataclass
class SetBool_Event(
    idl.IdlStruct,
    typename="std_srvs/srv/SetBool_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[SetBool_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[SetBool_Response, 1] = field(default_factory=list)


SetBool: idl.IdlServiceType[SetBool_Request, SetBool_Response, SetBool_Event] = (
    idl.make_idl_service(
        SetBool_Request,
        SetBool_Response,
        event_type=SetBool_Event,
    )
)

@dataclass
class Trigger_Request(idl.IdlStruct, typename="std_srvs/srv/Trigger_Request"):
    pass


@dataclass
class Trigger_Response(idl.IdlStruct, typename="std_srvs/srv/Trigger_Response"):
    success: bool = False
    message: str = ""


@dataclass
class Trigger_Event(
    idl.IdlStruct,
    typename="std_srvs/srv/Trigger_Event",
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[Trigger_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[Trigger_Response, 1] = field(default_factory=list)


Trigger: idl.IdlServiceType[Trigger_Request, Trigger_Response, Trigger_Event] = (
    idl.make_idl_service(
        Trigger_Request,
        Trigger_Response,
        event_type=Trigger_Event,
    )
)



import hashlib
from dataclasses import dataclass, field, fields
from typing import (
    Any,
    ClassVar,
    Final,
    Generic,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Self,
    Sequence,
    Type,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

import cyclonedds_idl as _idl

from .utils.description import (
    cyclonedds_struct_to_ros_type_description_json,
    ros2_type_hash_from_json,
)

types = _idl.types


# derive from the same metaclass Cyclone already uses
class IdlMetaIgnoreFinal(type(_idl.IdlStruct)):
    def __new__(mcls, name, bases, namespace, **kwargs):
        """
        Create a class while stripping ``Final`` and ``Literal`` field annotations.
        """
        ann = namespace.get("__annotations__")
        if ann:
            # remove only from annotations; keep the class attributes themselves
            for key in list(ann):
                if get_origin(ann[key]) is ClassVar:
                    ann.pop(key)
                elif get_origin(ann[key]) is Literal:
                    ann.pop(key)
                elif get_origin(ann[key]) is Final:
                    ann.pop(key)
        return super().__new__(mcls, name, bases, namespace, **kwargs)


class IdlStruct(_idl.IdlStruct, metaclass=IdlMetaIgnoreFinal):

    def serialize(
        self,
        buffer: Optional[_idl.Buffer] = None,
        endianness: Optional[_idl.Endianness] = None,
        use_version_2: Optional[bool] = None,
    ) -> bytes:
        """
        Serialize this IDL object, using a placeholder for truly empty messages.

        Args:
            buffer: Optional serialization buffer.
            endianness: Optional endianness override.
            use_version_2: Optional XCDR v2 override.

        Returns:
            Serialized bytes.
        """
        if len(getattr(type(self), "__annotations__", {})) == 0:
            # ROS 2 cannot send a truly empty message with only a header
            return DummyEmpty().serialize()
        return super().serialize(buffer, endianness, use_version_2)

    @classmethod
    def json_type_description(
        cls,
        *,
        root_type_name: str | None = None,
        type_name_overrides: Mapping[type, str] | None = None,
        indent: int = 2,
    ) -> str:
        """
        Build the ROS 2 type description JSON for this class.

        Args:
            root_type_name: Optional root ROS type name.
            type_name_overrides: Optional nested type name overrides.
            indent: JSON indentation level.

        Returns:
            Type description JSON.
        """
        return cyclonedds_struct_to_ros_type_description_json(
            cls,
            root_type_name=root_type_name,
            type_name_overrides=type_name_overrides,
            indent=indent,
        )

    @classmethod
    def _hash_rihs01_raw(cls) -> "hashlib._Hash":
        """
        Compute the raw RIHS01 hash object for this type.
        """
        return ros2_type_hash_from_json(cls.json_type_description())

    @classmethod
    def hash_rihs01(cls) -> str:
        """
        Compute the RIHS01 hash string for this type.
        """
        return f"RIHS01_{cls._hash_rihs01_raw().hexdigest()}"

    @classmethod
    def get_type_name(cls) -> str:
        """
        Return the ROS type name stored on the class.
        """
        return getattr(cls, "__idl_typename__", "")

    @classmethod
    def get_ros_type(cls) -> type:
        """
        Resolve the matching ROS Python type.
        """
        from importlib import import_module

        module_name, class_name = cls.get_type_name().replace("/", ".").rsplit(".", 1)
        mod = import_module(module_name)
        return getattr(mod, class_name)

    @classmethod
    def to_ros_type(cls) -> type:
        """
        Resolve the matching ROS Python type.
        """
        return cls.get_ros_type()

    def to_ros(self) -> object:
        """
        Convert this IDL object to the matching ROS message.
        """
        ros_msg = type(self).get_ros_type()()
        type_hints = get_type_hints(type(self))

        for f in fields(self):
            if not hasattr(ros_msg, f.name):
                continue

            src_type = type_hints.get(f.name, f.type)
            src_value = getattr(self, f.name)
            dst_value = getattr(ros_msg, f.name)
            setattr(ros_msg, f.name, self._to_ros_value(src_type, src_value, dst_value))

        return ros_msg

    @classmethod
    def from_ros(cls, msg: object) -> Self:
        """
        Convert a ROS message to this IDL type.

        Args:
            msg: Matching ROS message instance.

        Returns:
            Converted IDL instance.
        """
        if isinstance(msg, cls):
            return msg

        ros_type = cls.get_ros_type()
        if not isinstance(msg, ros_type):
            raise TypeError(
                f"{cls.__name__}.from_ros() expected {ros_type.__module__}.{ros_type.__name__}, "
                f"got {type(msg).__module__}.{type(msg).__name__}"
            )

        type_hints = get_type_hints(cls)
        kwargs: dict[str, Any] = {}

        for f in fields(cls):
            if not hasattr(msg, f.name):
                raise AttributeError(
                    f"ROS message {type(msg).__module__}.{type(msg).__name__} "
                    f"has no field {f.name!r}"
                )

            dst_type = type_hints.get(f.name, f.type)
            src_value = getattr(msg, f.name)
            kwargs[f.name] = cls._from_ros_value(dst_type, src_value)
        final = cls()
        for k, v in kwargs.items():
            setattr(final, k, v)
        return final

    @classmethod
    def _from_ros_value(cls, dst_type: Any, value: Any) -> Any:
        """
        Convert one ROS field value to its IDL-side value.

        Args:
            dst_type: Target field type.
            value: ROS field value.

        Returns:
            Converted Python value.
        """
        if value is None:
            return None

        # Nested ROS message -> nested IdlStruct
        if isinstance(dst_type, type) and issubclass(dst_type, IdlStruct):
            return dst_type.from_ros(value)

        # sequence[...] / array[...] / list[...] style annotations
        args = get_args(dst_type) or getattr(dst_type, "__args__", ())
        if (
            args
            and isinstance(value, Sequence)
            and not isinstance(value, (str, bytes, bytearray))
        ):
            elem_type = args[0]
            return [cls._from_ros_value(elem_type, item) for item in value]

        # bytes -> int
        if isinstance(value, bytes):
            return int.from_bytes(value)

        # Primitive / already compatible
        return value

    @classmethod
    def _to_ros_value(cls, src_type: Any, value: Any, dst_value: Any = None) -> Any:
        """
        Convert one IDL field value to its ROS-side value.

        Args:
            src_type: Source field type.
            value: IDL field value.
            dst_value: Current ROS destination value.

        Returns:
            Value for the ROS field.
        """
        if value is None:
            return None

        if isinstance(value, IdlStruct):
            return value.to_ros()

        args = get_args(src_type) or getattr(src_type, "__args__", ())
        if (
            args
            and isinstance(value, Sequence)
            and not isinstance(value, (str, bytes, bytearray))
        ):
            elem_type = args[0]
            if isinstance(elem_type, type) and issubclass(elem_type, IdlStruct):
                return [cls._to_ros_value(elem_type, item) for item in value]
            return value

        if isinstance(dst_value, bytes):
            return bytes([int(value)])

        if src_type in (float, types.float32, types.float64):
            return float(value)

        if src_type in (
            int,
            types.byte,
            types.char,
            types.int8,
            types.uint8,
            types.int16,
            types.uint16,
            types.int32,
            types.uint32,
            types.int64,
            types.uint64,
        ):
            return int(value)

        return value


@dataclass
class DummyEmpty(IdlStruct, typename="does/not/matter/empty"):
    structure_needs_at_least_one_member: types.uint8 = 0


@dataclass
class Time(IdlStruct, typename="builtin_interfaces/msg/Time"):
    sec: types.int32 = 0
    nanosec: types.uint32 = 0


@dataclass
class ServiceEventInfo(IdlStruct, typename="service_msgs/msg/ServiceEventInfo"):
    REQUEST_SENT: ClassVar[Literal[0]] = 0
    REQUEST_RECEIVED: ClassVar[Literal[1]] = 1
    RESPONSE_SENT: ClassVar[Literal[2]] = 2
    RESPONSE_RECEIVED: ClassVar[Literal[3]] = 3

    event_type: types.uint8 = 0
    stamp: Time = field(default_factory=lambda *_: Time())
    client_gid: types.array[types.uint8, 16] = field(
        default_factory=lambda *_: [0] * 16
    )
    sequence_number: types.int64 = 0


RequestT = TypeVar("RequestT", bound=IdlStruct)
ResponseT = TypeVar("ResponseT", bound=IdlStruct)


class IdlServiceEventStruct(IdlStruct, Generic[RequestT, ResponseT]):
    info: ServiceEventInfo
    request: types.sequence[RequestT, 1]
    response: types.sequence[ResponseT, 1]


EventT = TypeVar("EventT", bound=IdlStruct)


class IdlServiceStruct(IdlStruct, Generic[RequestT, ResponseT, EventT]):
    request_message: RequestT
    response_message: ResponseT
    event_message: EventT


class IdlServiceType(Protocol[RequestT, ResponseT, EventT]):
    Request: type[RequestT]
    Response: type[ResponseT]
    Event: type[EventT]

    def __call__(
        self,
        request_message: RequestT = ...,
        response_message: ResponseT = ...,
        event_message: EventT = ...,
    ) -> IdlServiceStruct[RequestT, ResponseT, EventT]: ...

    def get_type_name(self) -> str: ...
    def get_ros_type(self) -> type: ...
    def to_ros_type(self) -> type: ...
    def json_type_description(
        self,
        *,
        root_type_name: str | None = ...,
        type_name_overrides: Mapping[type, str] | None = ...,
        indent: int = ...,
    ) -> str: ...
    def hash_rihs01(self) -> str: ...


def _service_typename_from_message_types(
    request_type: type[RequestT], response_type: type[ResponseT]
) -> str:
    """
    Derive the shared service type name from request and response types.

    Args:
        request_type: Request message type.
        response_type: Response message type.

    Returns:
        Shared service type name.
    """
    request_typename = request_type.get_type_name()
    response_typename = response_type.get_type_name()

    if not request_typename.endswith("_Request"):
        raise ValueError(
            f"{request_type.__name__} must have a type name ending with '_Request', "
            f"got {request_typename!r}"
        )
    if not response_typename.endswith("_Response"):
        raise ValueError(
            f"{response_type.__name__} must have a type name ending with '_Response', "
            f"got {response_typename!r}"
        )

    service_typename = request_typename.removesuffix("_Request")
    expected_response_typename = f"{service_typename}_Response"
    if response_typename != expected_response_typename:
        raise ValueError(
            "Request and response types must share the same service type name, "
            f"got {request_typename!r} and {response_typename!r}"
        )
    return service_typename


def _make_service_event_type(
    *,
    service_typename: str,
    module_name: str,
    request_type: type[RequestT],
    response_type: type[ResponseT],
) -> type[IdlServiceEventStruct[RequestT, ResponseT]]:
    """
    Create the generated ``_Event`` type for a service.

    Args:
        service_typename: ROS service type name.
        module_name: Python module name.
        request_type: Request message type.
        response_type: Response message type.

    Returns:
        Generated event type.
    """
    service_name = service_typename.rsplit("/", 1)[1]
    event_name = f"{service_name}_Event"
    event_typename = f"{service_typename}_Event"
    event_namespace = {
        "__module__": module_name,
        "__idl_typename__": event_typename,
        "__annotations__": {
            "info": ServiceEventInfo,
            "request": types.sequence[request_type, 1],
            "response": types.sequence[response_type, 1],
        },
        "info": field(default_factory=ServiceEventInfo),
        "request": field(default_factory=list),
        "response": field(default_factory=list),
    }
    event_type = cast(
        type[IdlServiceEventStruct[RequestT, ResponseT]],
        dataclass(
            IdlMetaIgnoreFinal(
                event_name,
                (IdlServiceEventStruct,),
                event_namespace,
                typename=event_typename,
            )
        ),
    )
    event_type.__idl_typename__ = event_typename
    return event_type


AnyIdlServiceEvent: TypeAlias = IdlStruct
AnyIdlService: TypeAlias = IdlServiceStruct[IdlStruct, IdlStruct, IdlStruct]
AnyIdlServiceType: TypeAlias = IdlServiceType[IdlStruct, IdlStruct, IdlStruct]


SERVICE_FIELD_NAMES: Final[frozenset[str]] = frozenset(
    {"request_message", "response_message", "event_message"}
)


def is_service_type(obj: Any) -> TypeGuard[type[AnyIdlService]]:
    """
    Check whether an object looks like an IDL service type.
    """
    if not isinstance(obj, type):
        return False
    try:
        if not issubclass(obj, IdlStruct):
            return False
    except TypeError:
        return False

    dataclass_fields = getattr(obj, "__dataclass_fields__", {})
    return SERVICE_FIELD_NAMES.issubset(dataclass_fields)


@overload
def make_idl_service(
    request_type: type[RequestT],
    response_type: type[ResponseT],
    *,
    event_type: type[EventT],
    _module_name: str | None = None,
) -> IdlServiceType[RequestT, ResponseT, EventT]: ...


@overload
def make_idl_service(
    request_type: type[RequestT],
    response_type: type[ResponseT],
    *,
    _module_name: str | None = None,
) -> IdlServiceType[
    RequestT,
    ResponseT,
    IdlServiceEventStruct[RequestT, ResponseT],
]: ...


def make_idl_service(
    request_type: type[RequestT],
    response_type: type[ResponseT],
    *,
    event_type: type[EventT] | None = None,
    _module_name: str | None = None,
) -> (
    IdlServiceType[
        RequestT,
        ResponseT,
        EventT,
    ]
    | IdlServiceType[
        RequestT,
        ResponseT,
        IdlServiceEventStruct[RequestT, ResponseT],
    ]
):
    """
    Generate an IDL service type from request and response message types.

    Args:
        request_type: Request message type.
        response_type: Response message type.
        event_type: Optional explicit event message type.
        _module_name: Optional Python module name.

    Returns:
        Generated service type.
    """
    if not issubclass(request_type, IdlStruct):
        raise TypeError(
            f"request_type must inherit from IdlStruct, got {request_type!r}"
        )
    if not issubclass(response_type, IdlStruct):
        raise TypeError(
            f"response_type must inherit from IdlStruct, got {response_type!r}"
        )

    service_typename = _service_typename_from_message_types(request_type, response_type)
    service_name = service_typename.rsplit("/", 1)[1]
    module_name = _module_name or request_type.__module__
    uses_generated_event_type = event_type is None
    if event_type is None:
        resolved_event_type: type[IdlStruct] = _make_service_event_type(
            service_typename=service_typename,
            module_name=module_name,
            request_type=request_type,
            response_type=response_type,
        )
    else:
        if not issubclass(event_type, IdlStruct):
            raise TypeError(
                f"event_type must inherit from IdlStruct, got {event_type!r}"
            )
        expected_event_typename = f"{service_typename}_Event"
        if event_type.get_type_name() != expected_event_typename:
            raise ValueError(
                f"event_type must have type name {expected_event_typename!r}, "
                f"got {event_type.get_type_name()!r}"
            )
        resolved_event_type = event_type

    service_namespace = {
        "__module__": module_name,
        "__idl_typename__": service_typename,
        "__annotations__": {
            "request_message": request_type,
            "response_message": response_type,
            "event_message": resolved_event_type,
        },
        "Request": request_type,
        "Response": response_type,
        "Event": resolved_event_type,
        "request_message": field(default_factory=request_type),
        "response_message": field(default_factory=response_type),
        "event_message": field(default_factory=resolved_event_type),
    }
    service_type = dataclass(
        IdlMetaIgnoreFinal(
            service_name,
            (IdlServiceStruct,),
            service_namespace,
            typename=service_typename,
        )
    )
    service_type.__idl_typename__ = service_typename
    if uses_generated_event_type:
        return cast(
            IdlServiceType[
                RequestT,
                ResponseT,
                IdlServiceEventStruct[RequestT, ResponseT],
            ],
            service_type,
        )
    return cast(
        IdlServiceType[
            RequestT,
            ResponseT,
            EventT,
        ],
        service_type,
    )



from typing import get_type_hints

import pytest

from ros2_pyterfaces import idl
from ros2_pyterfaces.all_srvs import (
    SetBool_Event,
    SetBool_Request,
    SetBool_Response,
    Trigger_Response,
)
from ros2_pyterfaces.service_msgs.msg import ServiceEventInfo


SETBOOL_HASH = (
    "RIHS01_abe9e4bb6b41b40e6789712c00ec8871923e089af3f667a79992a428cff2da0a"
)


def test_make_idl_service_builds_expected_service_and_event_types():
    generated_type = idl.make_idl_service(
        SetBool_Request,
        SetBool_Response,
        _module_name=__name__,
    )

    service_hints = get_type_hints(generated_type)
    event_type = service_hints["event_message"]
    event_hints = get_type_hints(event_type, include_extras=True)

    assert issubclass(generated_type, idl.IdlStruct)
    assert issubclass(generated_type, idl.IdlServiceStruct)
    assert idl.is_service_type(generated_type)
    assert generated_type.get_type_name() == "std_srvs/srv/SetBool"

    assert generated_type.Request is SetBool_Request
    assert generated_type.Response is SetBool_Response
    assert generated_type.Event is event_type
    assert service_hints["request_message"] is SetBool_Request
    assert service_hints["response_message"] is SetBool_Response
    assert issubclass(event_type, idl.IdlStruct)
    assert issubclass(event_type, idl.IdlServiceEventStruct)
    assert event_type.get_type_name() == "std_srvs/srv/SetBool_Event"

    assert event_hints["info"] is ServiceEventInfo
    assert event_hints["request"] == idl.types.sequence[SetBool_Request, 1]
    assert event_hints["response"] == idl.types.sequence[SetBool_Response, 1]


def test_make_idl_service_uses_explicit_event_type():
    generated_type = idl.make_idl_service(
        SetBool_Request,
        SetBool_Response,
        event_type=SetBool_Event,
        _module_name=__name__,
    )

    service_hints = get_type_hints(generated_type)

    assert generated_type.Event is SetBool_Event
    assert service_hints["event_message"] is SetBool_Event
    assert generated_type.hash_rihs01() == SETBOOL_HASH


def test_make_idl_service_rejects_mismatched_request_response_types():
    with pytest.raises(ValueError, match="must share the same service type name"):
        idl.make_idl_service(SetBool_Request, Trigger_Response, _module_name=__name__)



from dataclasses import dataclass, field
from ..idl import IdlStruct, types
from ..geometry_msgs.msg import PoseStamped, Point, Pose, PoseWithCovariance, TwistWithCovariance, Accel, Twist, Wrench
from ..std_msgs.msg import Header
from ..builtin_interfaces.msg import Time

@dataclass
class Goals(IdlStruct, typename="nav_msgs/msg/Goals"):
    header: Header = field(default_factory=Header)
    goals: types.sequence[PoseStamped] = field(default_factory=list)


@dataclass
class GridCells(IdlStruct, typename="nav_msgs/msg/GridCells"):
    header: Header = field(default_factory=Header)
    cell_width: types.float32 = 0.0
    cell_height: types.float32 = 0.0
    cells: types.sequence[Point] = field(default_factory=list)


@dataclass
class MapMetaData(IdlStruct, typename="nav_msgs/msg/MapMetaData"):
    map_load_time: Time = field(default_factory=Time)
    resolution: types.float32 = 0.0
    width: types.uint32 = 0
    height: types.uint32 = 0
    origin: Pose = field(default_factory=Pose)


@dataclass
class Odometry(IdlStruct, typename="nav_msgs/msg/Odometry"):
    header: Header = field(default_factory=Header)
    child_frame_id: str = ""
    pose: PoseWithCovariance = field(default_factory=PoseWithCovariance)
    twist: TwistWithCovariance = field(default_factory=TwistWithCovariance)


@dataclass
class Path(IdlStruct, typename="nav_msgs/msg/Path"):
    header: Header = field(default_factory=Header)
    poses: types.sequence[PoseStamped] = field(default_factory=list)


@dataclass
class TrajectoryPoint(IdlStruct, typename="nav_msgs/msg/TrajectoryPoint"):
    header: Header = field(default_factory=Header)
    pose: Pose = field(default_factory=Pose)
    velocity: Twist = field(default_factory=Twist)
    acceleration: Accel = field(default_factory=Accel)
    effort: Wrench = field(default_factory=Wrench)


@dataclass
class OccupancyGrid(IdlStruct, typename="nav_msgs/msg/OccupancyGrid"):
    header: Header = field(default_factory=Header)
    info: MapMetaData = field(default_factory=MapMetaData)
    data: types.sequence[types.int8] = field(default_factory=list)


@dataclass
class Trajectory(IdlStruct, typename="nav_msgs/msg/Trajectory"):
    header: Header = field(default_factory=Header)
    points: types.sequence[TrajectoryPoint] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import ClassVar, Literal

import numpy as np

from ..builtin_interfaces.msg import Time
from ..geometry_msgs.msg import Point32, Quaternion, Transform, Twist, Vector3, Wrench
from ..idl import IdlStruct, types
from ..std_msgs.msg import Header


@dataclass
class BatteryState(IdlStruct, typename="sensor_msgs/msg/BatteryState"):
    POWER_SUPPLY_STATUS_UNKNOWN: ClassVar[Literal[0]] = 0
    POWER_SUPPLY_STATUS_CHARGING: ClassVar[Literal[1]] = 1
    POWER_SUPPLY_STATUS_DISCHARGING: ClassVar[Literal[2]] = 2
    POWER_SUPPLY_STATUS_NOT_CHARGING: ClassVar[Literal[3]] = 3
    POWER_SUPPLY_STATUS_FULL: ClassVar[Literal[4]] = 4
    POWER_SUPPLY_HEALTH_UNKNOWN: ClassVar[Literal[0]] = 0
    POWER_SUPPLY_HEALTH_GOOD: ClassVar[Literal[1]] = 1
    POWER_SUPPLY_HEALTH_OVERHEAT: ClassVar[Literal[2]] = 2
    POWER_SUPPLY_HEALTH_DEAD: ClassVar[Literal[3]] = 3
    POWER_SUPPLY_HEALTH_OVERVOLTAGE: ClassVar[Literal[4]] = 4
    POWER_SUPPLY_HEALTH_UNSPEC_FAILURE: ClassVar[Literal[5]] = 5
    POWER_SUPPLY_HEALTH_COLD: ClassVar[Literal[6]] = 6
    POWER_SUPPLY_HEALTH_WATCHDOG_TIMER_EXPIRE: ClassVar[Literal[7]] = 7
    POWER_SUPPLY_HEALTH_SAFETY_TIMER_EXPIRE: ClassVar[Literal[8]] = 8
    POWER_SUPPLY_TECHNOLOGY_UNKNOWN: ClassVar[Literal[0]] = 0
    POWER_SUPPLY_TECHNOLOGY_NIMH: ClassVar[Literal[1]] = 1
    POWER_SUPPLY_TECHNOLOGY_LION: ClassVar[Literal[2]] = 2
    POWER_SUPPLY_TECHNOLOGY_LIPO: ClassVar[Literal[3]] = 3
    POWER_SUPPLY_TECHNOLOGY_LIFE: ClassVar[Literal[4]] = 4
    POWER_SUPPLY_TECHNOLOGY_NICD: ClassVar[Literal[5]] = 5
    POWER_SUPPLY_TECHNOLOGY_LIMN: ClassVar[Literal[6]] = 6
    POWER_SUPPLY_TECHNOLOGY_TERNARY: ClassVar[Literal[7]] = 7
    POWER_SUPPLY_TECHNOLOGY_VRLA: ClassVar[Literal[8]] = 8
    header: Header = field(default_factory=Header)
    voltage: types.float32 = 0.0
    temperature: types.float32 = 0.0
    current: types.float32 = 0.0
    charge: types.float32 = 0.0
    capacity: types.float32 = 0.0
    design_capacity: types.float32 = 0.0
    percentage: types.float32 = 0.0
    power_supply_status: types.uint8 = 0
    power_supply_health: types.uint8 = 0
    power_supply_technology: types.uint8 = 0
    present: bool = False
    cell_voltage: types.sequence[types.float32] = field(default_factory=list)
    cell_temperature: types.sequence[types.float32] = field(default_factory=list)
    location: str = ""
    serial_number: str = ""


@dataclass
class ChannelFloat32(IdlStruct, typename="sensor_msgs/msg/ChannelFloat32"):
    name: str = ""
    values: types.sequence[types.float32] = field(default_factory=list)


@dataclass
class CompressedImage(IdlStruct, typename="sensor_msgs/msg/CompressedImage"):
    header: Header = field(default_factory=Header)
    format: str = ""
    data: types.sequence[types.uint8] = field(default_factory=list)


@dataclass
class FluidPressure(IdlStruct, typename="sensor_msgs/msg/FluidPressure"):
    header: Header = field(default_factory=Header)
    fluid_pressure: types.float64 = 0.0
    variance: types.float64 = 0.0


@dataclass
class Illuminance(IdlStruct, typename="sensor_msgs/msg/Illuminance"):
    header: Header = field(default_factory=Header)
    illuminance: types.float64 = 0.0
    variance: types.float64 = 0.0


@dataclass
class Image(IdlStruct, typename="sensor_msgs/msg/Image"):
    header: Header = field(default_factory=Header)
    height: types.uint32 = 0
    width: types.uint32 = 0
    encoding: str = ""
    is_bigendian: types.uint8 = 0
    step: types.uint32 = 0
    data: types.sequence[types.uint8] = field(default_factory=list)


@dataclass
class Imu(IdlStruct, typename="sensor_msgs/msg/Imu"):
    header: Header = field(default_factory=Header)
    orientation: Quaternion = field(default_factory=Quaternion)
    orientation_covariance: types.array[types.float64, 9] = field(
        default_factory=lambda: np.zeros(9)
    )
    angular_velocity: Vector3 = field(default_factory=Vector3)
    angular_velocity_covariance: types.array[types.float64, 9] = field(
        default_factory=lambda: np.zeros(9)
    )
    linear_acceleration: Vector3 = field(default_factory=Vector3)
    linear_acceleration_covariance: types.array[types.float64, 9] = field(
        default_factory=lambda: np.zeros(9)
    )


@dataclass
class JointState(IdlStruct, typename="sensor_msgs/msg/JointState"):
    header: Header = field(default_factory=Header)
    name: types.sequence[str] = field(default_factory=list)
    position: types.sequence[types.float64] = field(default_factory=list)
    velocity: types.sequence[types.float64] = field(default_factory=list)
    effort: types.sequence[types.float64] = field(default_factory=list)


@dataclass
class Joy(IdlStruct, typename="sensor_msgs/msg/Joy"):
    header: Header = field(default_factory=Header)
    axes: types.sequence[types.float32] = field(default_factory=list)
    buttons: types.sequence[types.int32] = field(default_factory=list)


@dataclass
class JoyFeedback(IdlStruct, typename="sensor_msgs/msg/JoyFeedback"):
    TYPE_LED: ClassVar[Literal[0]] = 0
    TYPE_RUMBLE: ClassVar[Literal[1]] = 1
    TYPE_BUZZER: ClassVar[Literal[2]] = 2
    type: types.uint8 = 0
    id: types.uint8 = 0
    intensity: types.float32 = 0.0


@dataclass
class LaserEcho(IdlStruct, typename="sensor_msgs/msg/LaserEcho"):
    echoes: types.sequence[types.float32] = field(default_factory=list)


@dataclass
class LaserScan(IdlStruct, typename="sensor_msgs/msg/LaserScan"):
    header: Header = field(default_factory=Header)
    angle_min: types.float32 = 0.0
    angle_max: types.float32 = 0.0
    angle_increment: types.float32 = 0.0
    time_increment: types.float32 = 0.0
    scan_time: types.float32 = 0.0
    range_min: types.float32 = 0.0
    range_max: types.float32 = 0.0
    ranges: types.sequence[types.float32] = field(default_factory=list)
    intensities: types.sequence[types.float32] = field(default_factory=list)


@dataclass
class MagneticField(IdlStruct, typename="sensor_msgs/msg/MagneticField"):
    header: Header = field(default_factory=Header)
    magnetic_field: Vector3 = field(default_factory=Vector3)
    magnetic_field_covariance: types.array[types.float64, 9] = field(
        default_factory=lambda: np.zeros(9)
    )


@dataclass
class MultiDOFJointState(IdlStruct, typename="sensor_msgs/msg/MultiDOFJointState"):
    header: Header = field(default_factory=Header)
    joint_names: types.sequence[str] = field(default_factory=list)
    transforms: types.sequence[Transform] = field(default_factory=list)
    twist: types.sequence[Twist] = field(default_factory=list)
    wrench: types.sequence[Wrench] = field(default_factory=list)


@dataclass
class NavSatStatus(IdlStruct, typename="sensor_msgs/msg/NavSatStatus"):
    STATUS_UNKNOWN: ClassVar[Literal[-2]] = -2
    STATUS_NO_FIX: ClassVar[Literal[-1]] = -1
    STATUS_FIX: ClassVar[Literal[0]] = 0
    STATUS_SBAS_FIX: ClassVar[Literal[1]] = 1
    STATUS_GBAS_FIX: ClassVar[Literal[2]] = 2
    SERVICE_UNKNOWN: ClassVar[Literal[0]] = 0
    SERVICE_GPS: ClassVar[Literal[1]] = 1
    SERVICE_GLONASS: ClassVar[Literal[2]] = 2
    SERVICE_COMPASS: ClassVar[Literal[4]] = 4
    SERVICE_GALILEO: ClassVar[Literal[8]] = 8
    status: types.int8 = -2
    service: types.uint16 = 0


@dataclass
class PointField(IdlStruct, typename="sensor_msgs/msg/PointField"):
    INT8: ClassVar[Literal[1]] = 1
    UINT8: ClassVar[Literal[2]] = 2
    INT16: ClassVar[Literal[3]] = 3
    UINT16: ClassVar[Literal[4]] = 4
    INT32: ClassVar[Literal[5]] = 5
    UINT32: ClassVar[Literal[6]] = 6
    FLOAT32: ClassVar[Literal[7]] = 7
    FLOAT64: ClassVar[Literal[8]] = 8
    INT64: ClassVar[Literal[9]] = 9
    UINT64: ClassVar[Literal[10]] = 10
    BOOL: ClassVar[Literal[11]] = 11
    name: str = ""
    offset: types.uint32 = 0
    datatype: types.uint8 = 0
    count: types.uint32 = 0


@dataclass
class Range(IdlStruct, typename="sensor_msgs/msg/Range"):
    ULTRASOUND: ClassVar[Literal[0]] = 0
    INFRARED: ClassVar[Literal[1]] = 1
    header: Header = field(default_factory=Header)
    radiation_type: types.uint8 = 0
    field_of_view: types.float32 = 0.0
    min_range: types.float32 = 0.0
    max_range: types.float32 = 0.0
    range: types.float32 = 0.0
    variance: types.float32 = 0.0


@dataclass
class RegionOfInterest(IdlStruct, typename="sensor_msgs/msg/RegionOfInterest"):
    x_offset: types.uint32 = 0
    y_offset: types.uint32 = 0
    height: types.uint32 = 0
    width: types.uint32 = 0
    do_rectify: bool = False


@dataclass
class RelativeHumidity(IdlStruct, typename="sensor_msgs/msg/RelativeHumidity"):
    header: Header = field(default_factory=Header)
    relative_humidity: types.float64 = 0.0
    variance: types.float64 = 0.0


@dataclass
class Temperature(IdlStruct, typename="sensor_msgs/msg/Temperature"):
    header: Header = field(default_factory=Header)
    temperature: types.float64 = 0.0
    variance: types.float64 = 0.0


@dataclass
class TimeReference(IdlStruct, typename="sensor_msgs/msg/TimeReference"):
    header: Header = field(default_factory=Header)
    time_ref: Time = field(default_factory=Time)
    source: str = ""


@dataclass
class CameraInfo(IdlStruct, typename="sensor_msgs/msg/CameraInfo"):
    header: Header = field(default_factory=Header)
    height: types.uint32 = 0
    width: types.uint32 = 0
    distortion_model: str = ""
    d: types.sequence[types.float64] = field(default_factory=list)
    k: types.array[types.float64, 9] = field(default_factory=lambda: np.zeros(9))
    r: types.array[types.float64, 9] = field(default_factory=lambda: np.zeros(9))
    p: types.array[types.float64, 12] = field(default_factory=lambda: np.zeros(12))
    binning_x: types.uint32 = 0
    binning_y: types.uint32 = 0
    roi: RegionOfInterest = field(default_factory=RegionOfInterest)


@dataclass
class JoyFeedbackArray(IdlStruct, typename="sensor_msgs/msg/JoyFeedbackArray"):
    array: types.sequence[JoyFeedback] = field(default_factory=list)


@dataclass
class MultiEchoLaserScan(IdlStruct, typename="sensor_msgs/msg/MultiEchoLaserScan"):
    header: Header = field(default_factory=Header)
    angle_min: types.float32 = 0.0
    angle_max: types.float32 = 0.0
    angle_increment: types.float32 = 0.0
    time_increment: types.float32 = 0.0
    scan_time: types.float32 = 0.0
    range_min: types.float32 = 0.0
    range_max: types.float32 = 0.0
    ranges: types.sequence[LaserEcho] = field(default_factory=list)
    intensities: types.sequence[LaserEcho] = field(default_factory=list)


@dataclass
class NavSatFix(IdlStruct, typename="sensor_msgs/msg/NavSatFix"):
    COVARIANCE_TYPE_UNKNOWN: ClassVar[Literal[0]] = 0
    COVARIANCE_TYPE_APPROXIMATED: ClassVar[Literal[1]] = 1
    COVARIANCE_TYPE_DIAGONAL_KNOWN: ClassVar[Literal[2]] = 2
    COVARIANCE_TYPE_KNOWN: ClassVar[Literal[3]] = 3
    header: Header = field(default_factory=Header)
    status: NavSatStatus = field(default_factory=NavSatStatus)
    latitude: types.float64 = 0.0
    longitude: types.float64 = 0.0
    altitude: types.float64 = 0.0
    position_covariance: types.array[types.float64, 9] = field(
        default_factory=lambda: np.zeros(9)
    )
    position_covariance_type: types.uint8 = 0


@dataclass
class PointCloud(IdlStruct, typename="sensor_msgs/msg/PointCloud"):
    header: Header = field(default_factory=Header)
    points: types.sequence[Point32] = field(default_factory=list)
    channels: types.sequence[ChannelFloat32] = field(default_factory=list)


@dataclass
class PointCloud2(IdlStruct, typename="sensor_msgs/msg/PointCloud2"):
    header: Header = field(default_factory=Header)
    height: types.uint32 = 0
    width: types.uint32 = 0
    fields: types.sequence[PointField] = field(default_factory=list)
    is_bigendian: bool = False
    point_step: types.uint32 = 0
    row_step: types.uint32 = 0
    data: types.sequence[types.uint8] = field(default_factory=list)
    is_dense: bool = False



import importlib
import pkgutil

import pytest


PACKAGES = [
    "ros2_pyterfaces",
]


def iter_submodules(package_name: str):
    """Yield a package/module and all its recursive submodules."""
    root = importlib.import_module(package_name)
    yield package_name

    if not hasattr(root, "__path__"):
        return

    seen = set()
    stack = [(root.__name__, list(root.__path__))]

    while stack:
        parent_name, parent_paths = stack.pop()

        for module_info in pkgutil.iter_modules(parent_paths, prefix=parent_name + "."):
            name = module_info.name
            if name in seen:
                continue
            seen.add(name)
            yield name

            if module_info.ispkg:
                subpkg = importlib.import_module(name)
                if hasattr(subpkg, "__path__"):
                    stack.append((subpkg.__name__, list(subpkg.__path__)))


@pytest.mark.parametrize("package_name", PACKAGES)
def test_recursive_imports(package_name: str):
    failures = []

    for module_name in sorted(iter_submodules(package_name)):
        try:
            importlib.import_module(module_name)
        except Exception as exc:
            failures.append(f"{module_name}: {type(exc).__name__}: {exc}")

    assert not failures, "Import failures:\n\n" + "\n".join(failures)



from .composition_interfaces.srv import *
from .diagnostic_msgs.srv import *
from .lifecycle_msgs.srv import *
from .nav_msgs.srv import *
from .rcl_interfaces.srv import *
from .sensor_msgs.srv import *
from .std_srvs.srv import *
from .type_description_interfaces.srv import *
from .visualization_msgs.srv import *



from dataclasses import dataclass, field

from .. import idl
from ..service_msgs.msg import ServiceEventInfo
from .msg import KeyValue, TypeDescription, TypeSource


@dataclass
class GetTypeDescription_Request(
    idl.IdlStruct, typename="type_description_interfaces/srv/GetTypeDescription_Request"
):
    type_name: str = ""
    type_hash: str = ""
    include_type_sources: bool = True


@dataclass
class GetTypeDescription_Response(
    idl.IdlStruct, typename="type_description_interfaces/srv/GetTypeDescription_Response"
):
    successful: bool = False
    failure_reason: str = ""
    type_description: TypeDescription = field(default_factory=TypeDescription)
    type_sources: idl.types.sequence[TypeSource] = field(default_factory=list)
    extra_information: idl.types.sequence[KeyValue] = field(default_factory=list)


@dataclass
class GetTypeDescription_Event(
    idl.IdlStruct,
    typename="type_description_interfaces/srv/GetTypeDescription_Event"
):
    info: ServiceEventInfo = field(default_factory=ServiceEventInfo)
    request: idl.types.sequence[GetTypeDescription_Request, 1] = field(default_factory=list)
    response: idl.types.sequence[GetTypeDescription_Response, 1] = field(
        default_factory=list
    )


GetTypeDescription: idl.IdlServiceType[
    GetTypeDescription_Request,
    GetTypeDescription_Response,
    GetTypeDescription_Event,
] = idl.make_idl_service(
    GetTypeDescription_Request,
    GetTypeDescription_Response,
    event_type=GetTypeDescription_Event,
)
